from typing import TYPE_CHECKING

from feeluown.gui.provider_ui import AbstractProviderUi
from feeluown.utils import aio

from .ui import LibraryRenderer
from .provider import provider

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class LocalProviderUi(AbstractProviderUi):

    def __init__(self, app: 'GuiApp'):
        self._app = app

    @property
    def provider(self):
        return provider

    def register_pages(self, route):
        route('/local')(show_provider)

    def login_or_go_home(self):
        self._app.browser.goto(uri='/local')


def show_provider(req):
    if hasattr(req, 'ctx'):
        app: 'GuiApp' = req.ctx['app']
    else:
        app = req  # Backward compatibility
    app.pl_uimgr.clear()
    # app.playlists.add(provider.playlists)

    app.ui.left_panel.my_music_con.hide()
    app.ui.left_panel.playlists_con.hide()

    aio.run_afn(app.ui.table_container.set_renderer,
                LibraryRenderer(provider.songs, provider.albums, provider.artists))



import logging
from typing import TYPE_CHECKING, Optional, TypeVar, Generic

from PyQt6.QtWidgets import QWidget, QVBoxLayout

from feeluown.i18n import t
from feeluown.library import (
    SupportsRecListDailyPlaylists,
    SupportsRecACollectionOfSongs,
    Collection,
    SupportsRecListDailySongs,
    Provider,
    SupportsRecACollectionOfVideos,
)
from feeluown.utils.aio import run_fn, gather, run_afn
from feeluown.gui.widgets.img_card_list import (
    VideoCardListView,
    VideoCardListModel,
    VideoCardListDelegate,
)
from feeluown.gui.widgets.song_minicard_list import SongMiniCardListModel
from feeluown.gui.widgets.selfpaint_btn import PlayButton
from feeluown.gui.helpers import BgTransparentMixin
from feeluown.gui.components.recommendation_panel import Panel, RecPlaylistsPanel
from feeluown.gui.components.recommendation_panel import create_song_list_view
from feeluown.gui.pages.template import render_scroll_area_view

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


async def render(req, **kwargs):
    await render_scroll_area_view(req, View)


class HBody(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)


class Overview(QWidget):
    def __init__(self, app: "GuiApp"):
        super().__init__(parent=None)
        self._app = app

        if self._app.ai is not None:
            from feeluown.gui.components.ai_radio_card import AIRadioCard

            self._ai_radio_card = AIRadioCard(self._app)
        else:
            self._ai_radio_card = None

        self._main_layout = QVBoxLayout(self)
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.setSpacing(10)
        if self._ai_radio_card is not None:
            self._main_layout.addWidget(self._ai_radio_card)

    async def render(self):
        if self._ai_radio_card is not None:
            await self._ai_radio_card.render()


P = TypeVar("P", bound=Provider)


class SongsBasePanel(Panel, Generic[P]):
    """
    Base panel class for show a list of songs.
    """

    def __init__(self, title: str, app: "GuiApp", provider: P):
        self._app = app
        self._provider = provider
        self._initial_row_count = 3
        self.songs_list_view = songs_list_view = create_song_list_view(
            self._app,
            self._initial_row_count,
        )
        pixmap = Panel.get_provider_pixmap(app, provider.identifier)
        super().__init__(title, songs_list_view, pixmap)

        # Add play_all_btn to the left of the header
        self.play_all_btn = PlayButton()
        # Insert the play button after the header
        self._h_layout.insertWidget(2, self.play_all_btn)

        # Connect buttons
        self.play_all_btn.clicked.connect(lambda: run_afn(self._play_all))
        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)

    def _show_more_or_less(self, checked):
        if checked:
            self.songs_list_view.set_fixed_row_count(self._initial_row_count)
        else:
            self.songs_list_view.set_fixed_row_count(-1)

    async def _play_all(self):
        songs = await run_fn(self.songs_list_view.model().get_reader().readall)
        self._app.playlist.set_models(songs, next_=True)
        self._app.player.resume()

    def set_reader(self, reader):
        songs_model = SongMiniCardListModel.create(reader, self._app)
        self.songs_list_view.setModel(songs_model)


class RecDailySongsPanel(SongsBasePanel[SupportsRecListDailySongs]):
    def __init__(self, app: "GuiApp", provider: SupportsRecListDailySongs):
        super().__init__(t("recommended-daily-playlist"), app, provider)

    async def render(self):
        songs = await run_fn(self._provider.rec_list_daily_songs)
        self.set_reader(songs)


class RecSongsPanel(SongsBasePanel[SupportsRecACollectionOfSongs]):
    def __init__(self, app: "GuiApp", provider: SupportsRecACollectionOfSongs):
        super().__init__(t("recommended-feelin-lucky"), app, provider)

    async def render(self):
        coll: Collection = await run_fn(self._provider.rec_a_collection_of_songs)
        self.set_reader(coll.models)
        self.header.setText(coll.name)


class RecVideosPanel(Panel):
    def __init__(self, app: "GuiApp", provider: SupportsRecACollectionOfVideos):
        self._app = app
        self._provider = provider
        self._initial_row_count = 2
        self.video_list_view = video_list_view = VideoCardListView(
            no_scroll_v=True, fixed_row_count=self._initial_row_count
        )
        video_list_view.setItemDelegate(
            VideoCardListDelegate(
                video_list_view,
                card_min_width=200,
            )
        )
        pixmap = Panel.get_provider_pixmap(app, provider.identifier)
        super().__init__(t("recommended-videos"), video_list_view, pixmap)

        # Connect the fold_unfold_btn to show more or less videos
        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)
        video_list_view.play_video_needed.connect(self._app.playlist.play_model)

    def _show_more_or_less(self, checked):
        if checked:
            self.video_list_view.set_fixed_row_count(self._initial_row_count)
        else:
            self.video_list_view.set_fixed_row_count(-1)

    async def render(self):
        coll = await run_fn(self._provider.rec_a_collection_of_videos)
        videos = coll.models
        if videos:
            model = VideoCardListModel.create(videos, self._app)
            self.video_list_view.setModel(model)
            self.header.setText(coll.name)
        else:
            self.header.setText(t("recommended-videos-missing"))
            self.video_list_view.hide()


class View(QWidget, BgTransparentMixin):
    def __init__(self, app: "GuiApp"):
        super().__init__(parent=None)
        self._app = app

        self._overview = Overview(app)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(20, 10, 20, 0)
        self._layout.setSpacing(20)
        self._layout.addWidget(self._overview)

    async def render(self):
        panels = []
        settings = self._app.config.NEW_HOMEPAGE_SETTINGS
        for content in settings.get("contents", []):
            name = content["name"]
            if name == "RecListDailySongs":
                panel = self._handle_rec_list_daily_songs(content)
                if panel is not None:
                    panels.append(panel)
            elif name == "RecListDailyPlaylists":
                panel = self._handle_rec_list_daily_playlists(content)
                if panel is not None:
                    panels.append(panel)
            elif name == "RecACollectionOfSongs":
                panel = self._handle_rec_a_collection_of_songs(content)
                if panel is not None:
                    panels.append(panel)
            elif name == "RecACollectionOfVideos":
                panel = self._handle_rec_a_collection_of_videos(content)
                if panel is not None:
                    panels.append(panel)
        for panel in panels:
            self._layout.addWidget(panel)
        gather(*([panel.render() for panel in panels] + [self._overview.render()]))

    def _handle_rec_a_collection_of_videos(self, content: dict) -> Optional[Panel]:
        source = content["provider"]
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecACollectionOfVideos):
            return RecVideosPanel(self._app, provider)
        logger.warning(
            f"Invalid homepage content: {content}, "
            f"provider {source} not found or not supported"
        )
        return None

    def _handle_rec_list_daily_songs(self, content: dict) -> Optional[Panel]:
        source = content["provider"]
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecListDailySongs):
            return RecDailySongsPanel(self._app, provider)
        logger.warning(
            f"Invalid homepage content: {content}, "
            f"provider {source} not found or not supported"
        )
        return None

    def _handle_rec_list_daily_playlists(self, content: dict) -> Optional[Panel]:
        source = content["provider"]
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecListDailyPlaylists):
            # Homepage can mix panels from multiple providers, keep provider icon
            # visible to disambiguate each section's source.
            return RecPlaylistsPanel(self._app, provider, show_icon=True)
        logger.warning(
            f"Invalid homepage content: {content}, "
            f"provider {source} not found or not supported"
        )
        return None

    def _handle_rec_a_collection_of_songs(self, content: dict) -> Optional[Panel]:
        source = content["provider"]
        provider = self._app.library.get(source)
        if isinstance(provider, SupportsRecACollectionOfSongs):
            return RecSongsPanel(self._app, provider)
        logger.warning(
            f"Invalid homepage content: {content}, "
            f"provider {source} not found or not supported"
        )
        return None



from dataclasses import dataclass


@dataclass
class SessionOptions:
    # Always make the human-friendly protocol version as the default.
    rpc_version: str = '2.0'     # RPC protocol version.
    pubsub_version: str = '1.0'  # Pubsub protocol version.


class Request:
    """fuo protocol request object"""
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    # FIXME: maybe add a property 'stdin_content' for Request.
    def __init__(self, cmd, cmd_args=None,
                 cmd_options=None, options=None,
                 has_heredoc=False, heredoc_word=None):
        self.cmd = cmd
        self.cmd_args = cmd_args or []
        self.cmd_options = cmd_options or {}

        self.options = options or {}

        self.has_heredoc = has_heredoc
        self.heredoc_word = heredoc_word

    def set_heredoc_body(self, body):
        assert self.has_heredoc is True and self.heredoc_word
        self.cmd_args = [body]


class Response:
    def __init__(self, ok=True, text='', req=None, options=None):
        self.code = 'OK' if ok else 'Oops'
        self.text = text
        self.options = options or {}

        self.req = req



from typing import Optional
import logging
import warnings
from collections import namedtuple


logger = logging.getLogger(__name__)


Field = namedtuple('Field', ('name', 'type_', 'default', 'desc', 'warn'))


class Config:
    """Configuration module

    Users can configure the values of various options in the rc file."""

    def __init__(self, name: str = 'config', parent: Optional['Config'] = None):
        object.__setattr__(self, '_name', name)
        object.__setattr__(self, '_parent', parent)
        object.__setattr__(self, '_fields', {})
        object.__setattr__(self, '_undeclared_fields', {})

    def __getattr__(self, name):
        # tips: Can't use getattr to get the value here,
        # otherwise it will cause an infinite loop
        if name in ('_fields', '_parent', '_name', '_undeclared_fields'):
            return object.__getattribute__(self, name)
        if name in self._fields:
            try:
                return object.__getattribute__(self, name)
            except AttributeError:
                return self._fields[name].default
        elif name in self._undeclared_fields:
            return self._undeclared_fields[name]

        # Requirement:
        #   User may define config like
        #       app.plugin.X = Y
        #   When 'plugin' is not installed, such config should not raise an error.
        # To achieve this, return a subconfig when accessing an undeclared key.
        logger.warning(f'Undeclared subconfig: {self.fullname}.{name}')
        tmpconfig = Config(name=name, parent=self)
        self._undeclared_fields[name] = tmpconfig
        return tmpconfig

    def __setattr__(self, name, value):
        if name in self._fields:
            field = self._fields[name]
            if field.warn is not None:
                warnings.warn(
                    'Config field({}): {}'.format(name, field.warn), stacklevel=2
                )
            # TODO: Validate the value type
            object.__setattr__(self, name, value)
        else:
            logger.warning(f'Assign to an undeclared config key: {self.fullname}.{name}')

    @property
    def fullname(self) -> str:
        if self._parent is None:
            return self._name
        return f'{self._parent.fullname}.{self._name}'

    def deffield(self, name, type_=None, default=None, desc='', warn=None):
        """Define a configuration field

        :param str name: the field name. It SHOULD be capitalized except the field
            refers to a sub-config.
        :param type_: feild type.
        :param default: default value for the field.
        :param desc: description for the field.
        :param warn: if field is deprecated, set a warn message.
        """
        if name not in self._fields:
            self._fields[name] = Field(
                name=name, type_=type_, default=default, desc=desc, warn=warn
            )
        else:
            raise ValueError('Field({}) is already defined.'.format(name))



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QListWidget, QListWidgetItem, QAbstractItemView, QFrame

from feeluown.player import Lyric


class LyricView(QListWidget):
    """Scollable lyrics list view.

    Two slots can be connected:
    1. on_line_changed
    2. on_lyric_changed
    """

    def __init__(self, parent):
        super().__init__(parent)

        self._lyric = None
        self._alignment = Qt.AlignmentFlag.AlignLeft
        self._highlight_font_size = 18

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWordWrap(True)

        self.currentItemChanged.connect(self.on_item_changed)

    def set_lyric(self, lyric: Lyric):
        self.clear()
        self._lyric = lyric
        if lyric is None:
            return

        for i, line in enumerate(lyric.lines):
            item = self._create_item(line)
            if i == lyric.current_index:
                self.setCurrentItem(item)
            self.addItem(item)

    def _create_item(self, line):
        item = QListWidgetItem(line)
        item.setTextAlignment(self._alignment)
        return item

    def clear(self):
        super().clear()
        self._lyric = None

    def on_line_changed(self, _):
        if self._lyric is None:
            return
        index = self._lyric.current_index
        if index is not None:
            item = self.item(index)
            self.setCurrentItem(item)
            self.scrollToItem(item, QAbstractItemView.ScrollHint.PositionAtCenter)
        else:
            self.setCurrentItem(None)

    def on_lyric_changed(self, lyric, _, *__):
        self.set_lyric(lyric)

    def on_item_changed(self, current, previous):
        self.reset_item(previous)
        if current:
            font = current.font()
            font.setPixelSize(self._highlight_font_size)
            current.setFont(font)

    def reset_item(self, item):
        if item:
            item.setFont(self.font())



import logging
from abc import ABCMeta, abstractmethod
from typing import (
    List,
    Generic,
    TypeVar,
    Optional,
    Callable,
    Tuple,
    Iterable,
    cast,
    Sequence,
    AsyncIterable,
)
from threading import Lock

logger = logging.getLogger(__name__)

__all__ = (
    'create_reader',
    'SequentialReader',
    'RandomSequentialReader',
    'Reader',
    'AsyncReader',
    # Below are deprecated.
    'RandomReader',
    'wrap',
)

T = TypeVar("T")


class ReaderException(Exception):
    pass


class CantReadAll(ReaderException):
    pass


class Reader(Generic[T], metaclass=ABCMeta):
    """Base reader class.

    This class is mainly designed for the following use case::

        Provider often splits resource which has large body to chunks, such as
        a large playlist with 10k songs, the client need to send several request
        to fetch the whole playlist. Generally, we call this design Pagination.
        Moreover, different provider has different pagination API. We want a
        unified API, so we create the Reader class.

    Note ``read_*`` method may raise *ANY* exception. In practice, the caller
    should know what exception is possible to happen. For example, for the
    upper use case, ProviderIOError is possible to happen and others are not
    supposed.
    """

    @property
    @abstractmethod
    def count(self) -> int:
        """Total number of objects reader can read."""

    @abstractmethod
    def read_range(self, start: int, end: int) -> List[T]:
        """Read objects in range [start, end)."""

    @abstractmethod
    def read(self, index) -> T:
        """Read object by index.

        :raises IndexError:
        """

    @abstractmethod
    def readall(self) -> List[T]:
        """Read all objects.

        :raises CantReadAll:

        .. versionchanged:: 3.8.10
           Raise CantReadAll instead of ReadFailed exception. ReadFailed inherits
           from ProviderIOError, but Reader and ProviderIOError should not be coupled.
        """

    @abstractmethod
    def _read_next(self) -> T:
        """Read next object. Only for internal usage."""

    def __iter__(self):
        return self

    def __next__(self) -> T:
        """

        .. versionchanged:: 3.8.10
           Do not handle any exception here. The caller should be responsible to
           handle exceptions.
        """
        return self._read_next()


class SequentialReader(Reader[T]):
    """Help you sequential read data

    We only want to launch web request when we need the resource
    Formerly, we use Python generator to achieve this lazy read
    feature. However, we can't extract any read meta info,
    such as total count and current offset, from the ordinary
    generator.

    SequentialReader implements the iterator protocol, wraps the
    generator and store the reader state.

    .. note::

         iterating may be a blocking operation.

    **Usage example**:

    >>> def fetch_songs(page=1, page_size=50):
    ...     return list(range(page * page_size,
    ...                       (page + 1) * page_size))
    ...
    >>> def create_songs_g():
    ...     page = 0
    ...     total_page = 2
    ...     page_size = 2
    ...
    ...     def g():
    ...         nonlocal page, page_size
    ...         while page < total_page:
    ...            for song in fetch_songs(page, page_size):
    ...                yield song
    ...            page += 1
    ...
    ...     total = total_page * page_size
    ...     return SequentialReader(g(), total)
    ...
    >>> g = create_songs_g()
    >>> g.offset, g.count
    (0, 4)
    >>> next(g), next(g)
    (0, 1)
    >>> list(g)
    [2, 3]
    >>> g.offset, g.count
    (4, 4)

    .. versionadded:: 3.1
    """

    def __init__(self, g, count: Optional[int], offset: int = 0):
        """init

        :param g: Python generator
        :param offset: current offset
        :param count: total count. count can be None, which means the
                      total count is unknown. When it is unknown, be
                      CAREFUL to use list(reader).
        """
        super().__init__()
        self._g = g
        self._count = count
        self.offset = offset
        self._objects: List[T] = []
        self._lock = Lock()

    @property
    def count(self):
        return self._count

    def readall(self) -> List[T]:
        if self._count is None:
            raise CantReadAll("can't readall when count is unknown")
        list(self)
        return self._objects

    def read_range(self, start, end) -> List[T]:
        assert 0 <= start < end
        while len(self._objects) < end:
            try:
                next(self)
            except StopIteration:
                break
        return self._objects[start:end]

    def read(self, index):
        self.read_range(index, index+1)
        return self._objects[index]

    def _read_next(self) -> T:
        if self._count is None or self.offset < self.count:
            try:
                with self._lock:
                    obj = next(self._g)
            except StopIteration:
                if self._count is None:
                    self._count = self.offset
                raise
        else:
            raise StopIteration
        self.offset += 1
        self._objects.append(obj)
        return obj


class RandomSequentialReader(Reader[T]):

    def __init__(self,
                 count,
                 read_func: Callable[[int, int], Iterable[T]],
                 max_per_read=100):
        """random reader constructor

        :param int count: total number of objects
        :param function read_func: func(start: int, end: int) -> list
        :param int max_per_read: max count per read, it must big than 0
        """
        self.offset = 0
        self._count = count
        self._ranges: List[Tuple[int, int]] = []  # list of tuple
        self._objects: List[Optional[T]] = [None] * count
        self._read_func = read_func
        self._lock = Lock()

        assert max_per_read > 0, 'max_per_read must big than 0'
        self._max_per_read = max_per_read

    @property
    def count(self):
        return self._count

    def read(self, index):
        """read object by index

        If the object is not already read, this method may trigger IO operation.
        """
        yes, r = self._has_index(index)
        if yes:
            return self._objects[index]
        self._read_range(*r)
        return self._objects[index]

    def readall(self) -> List[T]:
        """read all objects

        :return list: list of objects
        :raises ReadFailed:
        """
        # all objects have been read
        if len(self._ranges) == 1 and self._ranges[0] == (0, self._count):
            return cast(List[T], self._objects)

        start = 0
        end = 0
        count = self._count
        while end < count:
            end = min(count, end + self._max_per_read)
            self._read_range(start, end)
            start = end
        return cast(List[T], self._objects)

    # def explain_readall(self):
    #     read_times = self._count / self._max_per_read
    #     if self._count % self._max_per_read > 0:
    #         read_times += 1
    #     return {'count': self._count,
    #             'max_per_read': self._max_per_read,
    #             'read_times': read_times}

    def read_range(self, start, end) -> List[T]:
        self._read_range(start, end)
        return cast(List[T], self._objects[start:end])

    def _read_range(self, start, end):
        # Though this method is thread safe now, _read_range_unsafe may read
        # the same range multiple times.
        with self._lock:
            self._read_range_unsafe(start, end)

    def _read_range_unsafe(self, start, end):
        assert start <= end, 'start should less than end'
        logger.debug('trigger read_func(%d, %d)', start, end)
        objs = list(self._read_func(start, end))
        actual = len(objs)
        self._objects[start:start + actual] = objs
        self._refresh_ranges()

    def _has_index(self, index):
        has_been_read = False
        left_index = right_index = None  # [left, right) -> range to read

        gt_index = None
        for r in self._ranges:
            start, end = r
            if index < start:
                # [, gt_index) index [start, end)
                gt_index = gt_index if gt_index is not None else 0
                left_index = index
                right_index = min(start, index + self._max_per_read)
                # trick: read as much as possible at a time to improve performance
                if start - gt_index <= self._max_per_read:
                    left_index = gt_index
                    right_index = start
                break
            # found index
            elif start <= index < end:
                has_been_read = True
                left_index, right_index = start, end
                break
            else:
                gt_index = end
        else:
            # default read range [index, min(index + max_per_read, self._count))
            left_index = index
            right_index = min(index + self._max_per_read, self._count)
            # trick: read as much as possible at a time to improve performance
            if gt_index is not None:
                if self._count - gt_index < self._max_per_read:
                    left_index = gt_index
                    right_index = self._count
        return has_been_read, (left_index, right_index)

    def _refresh_ranges(self):
        ranges = []
        start = None
        for i, obj in enumerate(self._objects):
            if start is None and obj is not None:
                start = i
                continue
            if start is not None and obj is None:
                ranges.append((start, i))
                start = None
        if start is not None:
            ranges.append((start, len(self._objects)))
        self._ranges = ranges

    def _read_next(self):
        if self.offset >= self._count:
            raise StopIteration
        obj = self.read(self.offset)
        self.offset += 1
        return obj


RandomReader = RandomSequentialReader  # For backward compatibility.


class AsyncReader:
    """Async version of reader.

    .. versionadded:: 3.8.10
    """

    pass


class AsyncSequentialReader(AsyncReader):

    def __init__(self, g, count, offset=0):
        """init

        :param g: Python generator
        :param offset: current offset
        :param count: total count. count can be None, which means the
                      total count is unknown. When it is unknown, be
                      CAREFUL to use list(reader).
        """
        self._g = g
        self._count = count
        self.offset = offset
        self._objects = []

    @property
    def count(self):
        return self._count

    async def a_readall(self):
        if self._count is None:
            raise CantReadAll("can't readall when count is unknown")
        async for _ in self:
            pass
        return self._objects

    async def a_read_next(self):
        if self._count is None or self.offset < self.count:
            try:
                obj = await self._g.asend(None)
            except StopAsyncIteration:
                if self._count is None:
                    self._count = self.offset + 1
                raise
        else:
            raise StopAsyncIteration
        self.offset += 1
        self._objects.append(obj)
        return obj

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return await self.a_read_next()
        except StopAsyncIteration:
            raise


def wrap(iterable):
    """

    .. versionadded:: 3.4
    .. deprecated:: 3.7.7
       Use :func:`create_reader` instead.
    """
    # if it is a reader already, just return it
    if isinstance(iterable, Reader):
        return iterable

    # async reader
    if isinstance(iterable, AsyncIterable):
        return AsyncSequentialReader(iterable, count=None)

    if not isinstance(iterable, Iterable):
        raise TypeError(f'must be a Iterable, got {type(iterable)}')
    if isinstance(iterable, Sequence):
        count = len(iterable)
        return RandomSequentialReader(count,
                                      lambda start, end: iterable[start:end],
                                      max_per_read=max(count, 1))
    # maybe a generator/iterator
    return SequentialReader(iterable, count=None)


def create_reader(iterable):
    """Create a reader from an iterable.

    >>> reader = wrap([1, 2])
    >>> isinstance(reader, RandomSequentialReader)
    True
    >>> reader.readall()
    [1, 2]
    >>> isinstance(wrap(iter([])), SequentialReader)
    True
    >>> wrap(None)
    Traceback (most recent call last):
        ...
    TypeError: must be a Iterable, got <class 'NoneType'>

    .. versionadded:: 3.7.7
    """
    return wrap(iterable)



import logging

from PyQt6.QtWidgets import (
    QLabel, QVBoxLayout, QHBoxLayout, QFrame
)

from feeluown.app.gui_app import GuiApp
from feeluown.ai import AISongMatcher
from feeluown.library import reverse
from feeluown.gui.widgets import PlayButton
from feeluown.gui.widgets.cover_label import CoverLabelV2
from feeluown.gui.widgets.header import LargeHeader
from feeluown.gui.helpers import ClickableMixin, random_solarized_color
from feeluown.utils import aio

logger = logging.getLogger(__name__)


class Header(ClickableMixin, LargeHeader):
    def __init__(self, text, **kwargs):
        super().__init__(text=text, **kwargs)


class AIRadioCard(QFrame):
    def __init__(self, app: GuiApp, parent=None):
        super().__init__(parent)
        self._app = app
        self.ai_radio = self._app.ai.get_copilot()

        # Widgets and layouts
        self._header = Header('AI 电台')
        self._header.setToolTip('暂未实现，欢迎 PR :)')
        self._summary_label = QLabel(self)
        self._summary_label.setWordWrap(True)
        self._cover_label = CoverLabelV2(self._app, self)
        self._status_label = QLabel(self)
        self._play_btn = PlayButton(length=20, padding=0.2)
        self._play_btn.setDisabled(True)
        self._play_btn.clicked.connect(self.on_play_btn_clicked)
        self._header.clicked.connect(self.on_header_clicked)
        self.setup_ui()

        self._song = None  # The song which is displayed.

    def setup_ui(self):
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(10, 10, 10, 10)
        self._layout.setSpacing(8)
        self.setFixedSize(360, 140)
        self._cover_label.setFixedSize(120, 120)
        self.setAutoFillBackground(True)
        self.setStyleSheet('border-radius: 3px; background: palette(window);')
        self._status_label.setStyleSheet('font-size: 15px;')
        self._header.setStyleSheet(
            f"QLabel:hover {{color: '{random_solarized_color().name()}'}};"
        )

        self._label_layout = QVBoxLayout()
        self._label_layout.addWidget(self._header)
        self._label_layout.addWidget(self._summary_label)
        self._label_layout.addStretch(0)
        self._song_layout = QHBoxLayout()
        self._song_layout.addWidget(self._play_btn)
        self._song_layout.addWidget(self._status_label)
        self._label_layout.addLayout(self._song_layout)
        self._layout.addLayout(self._label_layout)
        self._layout.addWidget(self._cover_label)

    async def render(self):
        self._status_label.setText('正在获取电台歌曲...')
        ai_songs = await self.ai_radio.recommend_a_song()
        if not ai_songs:
            self._status_label.setText('AI 推荐歌曲失败...')
            logger.warning("AI radio recommend_a_song return empty")
        else:
            first = ai_songs[0]
            self._status_label.setText(f'{first.title} • {first.artists_name}')
            self._summary_label.setText(first.description)
            song = await AISongMatcher(self._app).match(first)
            logger.info(f"matched song: {song}, ai_song: {first}")
            if song is not None:
                self._song = song
                self._play_btn.setEnabled(True)
                self._play_btn.setToolTip('点击播放')
                self._status_label.setText(f'{song.title} • {song.artists_name}')
                cover_media = await aio.run_fn(
                    self._app.library.model_get_cover_media, song
                )
                await self._cover_label.show_cover_media(cover_media, reverse(song))
            else:
                self._play_btn.setToolTip('未匹配到音源')

    def on_play_btn_clicked(self):
        self._app.playlist.play_model(self._song)

    def on_header_clicked(self):
        # TODO: to be implemented
        pass



# pylint: disable=duplicate-code
# FIXME: remove duplicate code lator
from typing import TYPE_CHECKING

from feeluown.library import SupportsToplist
from feeluown.gui.page_containers.table import TableContainer, Renderer
from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.utils.aio import run_fn
from feeluown.utils.reader import create_reader
from feeluown.i18n import t
from .template import render_error_message


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


async def render(req, **_):
    app: "GuiApp" = req.ctx["app"]
    pvd_ui = app.current_pvd_ui_mgr.get()
    if pvd_ui is None:
        return await render_error_message(app, t("provider-unknown-cannot-view"))

    provider = pvd_ui.provider
    scroll_area = ScrollArea()
    body = TableContainer(app, scroll_area)
    scroll_area.setWidget(body)
    app.ui.right_panel.set_body(scroll_area)
    if isinstance(provider, SupportsToplist):
        playlists = await run_fn(provider.toplist_list)
        renderer = Renderer()
        await body.set_renderer(renderer)
        renderer.show_playlists(create_reader(playlists))
        renderer.meta_widget.show()
        renderer.meta_widget.title = t("top-list")



from collections import defaultdict
import logging

from feeluown.serializers import serialize
from feeluown.server.protocol import DeadSubscriber


logger = logging.getLogger(__name__)


class Gateway:
    def __init__(self):
        self.topics = set()
        self._relations = defaultdict(set)  # {'topic': subscriber_set}

    def add_topic(self, topic):
        self.topics.add(topic)

    def remove_topic(self, topic):
        if topic in self.topics:
            self.topics.remove(topic)

    def link(self, topic, subscriber):
        self._relations[topic].add(subscriber)

    def unlink(self, topic, subscriber):
        if topic in self.topics and subscriber in self._relations[topic]:
            self._relations[topic].remove(subscriber)

    def remove_subscriber(self, subscriber):
        for topic in self.topics:
            if subscriber in self._relations[topic]:
                self._relations[topic].remove(subscriber)

    def publish(self, obj, topic, need_serialize=False):
        # NOTE: use queue? maybe.
        subscribers = self._relations[topic]
        for subscriber in subscribers.copy():
            try:
                if need_serialize is True:
                    msg = serialize('json', obj, brief=False)
                else:
                    msg = obj
                subscriber.write_topic_msg(topic, msg)
            except DeadSubscriber:
                # NOTE: need lock?
                self.remove_subscriber(subscriber)



#!/usr/bin/env python3

import asyncio
import json
import os
import socket
import sys
import subprocess
import time
import traceback
from contextlib import contextmanager

from feeluown.cli import Client, Request


@contextmanager
def create_client():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(('localhost', 23333))
    client = Client(sock)
    try:
        yield client
    finally:
        client.close()


def wait_until_23333_service_ok(timeout):
    i = 0
    while i < timeout:
        try:
            with create_client() as client:
                pass
        except ConnectionRefusedError:
            time.sleep(1)
            i += 1
        else:
            return True
    return False


def collect():
    for key in globals():
        if key.startswith('test_'):
            test_method = globals()[key]
            yield test_method


def test_show_providers_with_json_format():
    with create_client() as client:
        resp = asyncio.run(
            client.send(Request('show', ['fuo://'], options={'format': 'json'})))
        json.loads(resp.text)


def test_cmd_options():
    p = subprocess.run(['fuo', 'search', 'xx', '--type=album', '--source=dummy'])
    assert p.returncode == 0


def test_sub_live_lyric():
    for topic in ('topic.live_lyric', 'live_lyric'):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect(('127.0.0.1', 23334))
        sock.settimeout(1)
        welcome = sock.recv(100)
        assert 'pubsub' in welcome.decode('utf-8')
        sock.send(bytes('sub {}\r\n'.format(topic), 'utf-8'))
        response = sock.recv(100)
        assert 'OK' in response.decode('utf-8')
        sock.close()


def run():
    popen = subprocess.Popen(['fuo', '-v'])

    assert wait_until_23333_service_ok(timeout=10)

    failed = False
    for case in collect():
        print('{}...'.format(case.__name__), end='')
        try:
            case()
        except Exception as e:  # noqa
            print('failed')
            traceback.print_exc()
            failed = True
        else:
            print('ok')

    subprocess.run(['fuo', 'exec', 'app.exit()'])
    popen.wait(timeout=10)

    exists = os.path.exists(os.path.expanduser('~/.FeelUOwn/data/state.json'))
    # Since the app may crash during process terminating, the app is considered as
    # existing successfully when the state.json is saved.
    if not exists:
        print('app exits abnormally')
        failed = True

    returncode = 1 if failed is True else 0
    sys.exit(returncode)


if __name__ == '__main__':
    run()



import argparse
import textwrap
from functools import partial

from feeluown.i18n import t


def create_fmt_parser():
    fmt_parser = argparse.ArgumentParser(add_help=False)
    fmt_parser.add_argument(
        "--format",
        help="change command output format (default: plain)",
        choices=["plain", "json"],
    )
    fmt_parser.add_argument(
        "--json",
        action="store_true",
        default=False,
        help="change command output format to json",
    )
    fmt_parser.add_argument(
        "--plain",
        action="store_true",
        default=False,
        help="change command output format to plan",
    )
    return fmt_parser


def add_common_cmds(subparsers: argparse._SubParsersAction):
    fmt_parser = create_fmt_parser()
    parents = [fmt_parser]

    add_parser = partial(subparsers.add_parser, parents=parents)

    # Create parsers.
    #
    play_parser = add_parser(
        "play",
        description=textwrap.dedent("""\
        Example:
            - fuo play fuo://netease/songs/3027393
            - fuo play "in the end"
            - fuo play 稻香-周杰伦
        """),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    show_parser = add_parser(
        "show",
        help="显示资源详细信息",
    )
    search_parser = add_parser(
        "search",
        description=textwrap.dedent("""\
        Example:
            - fuo search hero
            - fuo search 李宗盛 source=xiami,type=artist
            - fuo search 李宗盛 [source=xiami,type=artist]
            - fuo search lizongsheng "source='xiami,qq',type=artist"
            - fuo search 李宗盛 "[source='xiami,qq',type=artist]"
        """),
        formatter_class=argparse.RawTextHelpFormatter,
        parents=parents,
    )
    remove_parser = add_parser("remove", help=t("command-track-remove"))
    add_parser_ = add_parser("add", help=t("command-track-add"))
    exec_parser = add_parser("exec")
    jsonrpc_parser = add_parser("jsonrpc")
    add_parser("pause", help=t("command-playback-pause"))
    add_parser("resume", help=t("command-playback-resume"))
    add_parser("toggle", help="")
    add_parser("stop")
    add_parser("next")
    add_parser("previous")
    add_parser("list")
    add_parser("clear")
    add_parser("status")

    # Initialize parsers.
    #
    play_parser.add_argument("uri")
    show_parser.add_argument("uri")
    remove_parser.add_argument("uri")
    add_parser_.add_argument("uri", nargs="?")
    search_parser.add_argument("keyword", help=t("command-search"))
    search_parser.add_argument("-s", "--source", action="append")
    search_parser.add_argument(
        "--type",
        action="append",
        type=str,
        choices=["song", "album", "artist", "video", "playlist"],
    )
    exec_parser.add_argument("code", nargs="?", help=t("command-exec-code"))
    jsonrpc_parser.add_argument("body", nargs="?", help=t("command-jsonrpc-body"))


def add_cli_cmds(subparsers: argparse._SubParsersAction):
    # Generate desktop file/icon.
    subparsers.add_parser("genicon", description="generate desktop icon")


def add_server_cmds(subparsers: argparse._SubParsersAction, include_pubsub=True):
    fmt_parser = create_fmt_parser()
    subparsers.add_parser(
        "quit",
        parents=[fmt_parser],
    )
    help_parser = subparsers.add_parser(
        "help",
        parents=[fmt_parser],
    )
    help_parser.add_argument("cmdname")
    # The default value of these values should be None.
    # See set handler for more details.
    set_parser = subparsers.add_parser(
        "set",
        help="设置会话参数",
        parents=[fmt_parser],
    )
    set_parser.add_argument(
        "--rpc-version",
        type=str,
        choices=["1.0", "2.0"],
    )
    set_parser.add_argument(
        "--pubsub-version",
        type=str,
        choices=["1.0", "2.0"],
    )

    if include_pubsub is True:
        sub_parser = subparsers.add_parser(
            "sub",
            help="订阅主题消息",
            parents=[fmt_parser],
        )
        sub_parser.add_argument(
            "topics",
            nargs="?",
            help="主题名字",
        )


def _create_cli_parser() -> argparse.ArgumentParser:
    # pylint: disable=import-outside-toplevel
    from feeluown import __version__ as feeluown_version

    parser = argparse.ArgumentParser(
        description=textwrap.dedent(
            """\
        FeelUOwn - modern music player (daemon).

        Example:
            - fuo                        # start fuo server
            - fuo status                 # lookup server status
            - fuo play 晴天-周杰伦       # search and play
        """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
        prog="feeluown",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {}".format(feeluown_version),
    )
    parser.add_argument(
        "-ns",
        "--no-server",
        action="store_true",
        default=False,
        help=t("cli-no-server"),
    )
    parser.add_argument(
        "-nw",
        "--no-window",
        action="store_true",
        default=False,
        help=t("cli-no-window"),
    )

    # options about log
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help=t("cli-debug-mode"),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help=t("cli-verbose"),
    )
    parser.add_argument(
        "--log-to-file",
        action="store_true",
        default=False,
        help=t("cli-log-to-file"),
    )

    # XXX: Not sure if it's possible to add a regex-based option?
    # For example, a `--mpv-*` option, otherwise I’d need to write
    # an option for each mpv configuration?

    # TODO: Need to include in the documentation
    # how to view the available playback devices.
    parser.add_argument(
        "--mpv-audio-device",
        help=t("cli-mpv-audio-device"),
    )
    return parser


def create_cli_parser() -> argparse.ArgumentParser:
    parser = _create_cli_parser()
    subparsers = parser.add_subparsers(
        dest="cmd",
    )

    add_common_cmds(subparsers)
    add_cli_cmds(subparsers)
    return parser



import logging
from typing import TYPE_CHECKING, Optional

from PyQt6.QtWidgets import QWidget, QVBoxLayout, QGridLayout, QSizePolicy
from feeluown.library.provider_protocol import SupportsToplist

from feeluown.library import (
    CollectionType,
    SupportsRecListDailyPlaylists,
    SupportsRecListDailySongs,
    SupportsRecListCollections,
    SupportsCurrentUserDislikeSongsReader,
    SupportsCurrentUserListRadioSongs,
)

from feeluown.gui.widgets import CalendarButton, RankButton, EmojiButton
from feeluown.gui.helpers import BgTransparentMixin
from feeluown.gui.components.recommendation_panel import (
    Panel,
    RecAlbumsCollectionPanel,
    RecPlaylistsPanel,
    RecPlaylistsCollectionPanel,
    RecSongsCollectionPanel,
    RecVideosCollectionPanel,
)
from feeluown.gui.pages.template import render_scroll_area_view
from feeluown.utils.aio import run_fn
from feeluown.i18n import t


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


logger = logging.getLogger(__name__)

# Keep action controls visually comparable with the recommendation panel title/content.
ActionButtonHeight = 34
ActionButtonMinWidth = 170
ActionButtonSpacing = 10
RecCollectionsLimit = 12


async def render(req, **kwargs):
    await render_scroll_area_view(req, View)


class View(QWidget, BgTransparentMixin):
    def __init__(self, app: "GuiApp"):
        super().__init__(parent=None)
        self._app = app
        self._heart_radar_provider: Optional[SupportsCurrentUserListRadioSongs] = None
        self._playlist_panel: Optional[RecPlaylistsPanel] = None
        self._collection_panels: list[Panel] = []

        self.daily_songs_btn = CalendarButton(
            t("recommended-daily-playlist"),
            height=ActionButtonHeight,
            parent=self,
        )
        self.rank_btn = RankButton(height=ActionButtonHeight, parent=self)
        # FIXME: design a new button for dislike
        self.dislike_btn = EmojiButton(
            "🚫", t("music-blacklisted"), height=ActionButtonHeight, parent=self
        )
        self.heart_radar_btn = EmojiButton(
            "📻", t("music-radio-radar"), height=ActionButtonHeight, parent=self
        )
        # Buttons should have comparable visual weight with the panel title/body.
        self.daily_songs_btn.setMinimumWidth(ActionButtonMinWidth)
        self.rank_btn.setMinimumWidth(ActionButtonMinWidth)
        self.dislike_btn.setMinimumWidth(ActionButtonMinWidth)
        self.heart_radar_btn.setMinimumWidth(ActionButtonMinWidth)
        self._action_btns = [
            self.daily_songs_btn,
            self.rank_btn,
            self.heart_radar_btn,
            self.dislike_btn,
        ]
        # Preserve semantic order when wrapping to 2/1 columns.
        for btn in self._action_btns:
            btn.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self._action_cols = 0

        self._recommendation_section = QWidget(self)
        self._recommendation_section_layout = QVBoxLayout(self._recommendation_section)
        self._recommendation_section_layout.setContentsMargins(0, 0, 0, 0)
        self._recommendation_section_layout.setSpacing(0)
        self._recommendation_section.hide()

        self._layout = QVBoxLayout(self)
        self._setup_ui()

        self.daily_songs_btn.clicked.connect(
            lambda: self._app.browser.goto(page="/rec/daily_songs")
        )
        self.rank_btn.clicked.connect(lambda: self._app.browser.goto(page="/toplist"))
        self.heart_radar_btn.clicked.connect(self._on_heart_radar_clicked)
        self.dislike_btn.clicked.connect(
            lambda: self._app.browser.goto(page="/my_dislike")
        )

    def _setup_ui(self):
        self._action_layout = QGridLayout()
        self._action_layout.setContentsMargins(0, 0, 0, 0)
        self._action_layout.setHorizontalSpacing(ActionButtonSpacing)
        self._action_layout.setVerticalSpacing(ActionButtonSpacing)

        self._layout.setContentsMargins(20, 10, 20, 0)
        self._layout.setSpacing(0)
        # Keep discovery action buttons as the topmost section to reduce noise.
        self._layout.addLayout(self._action_layout)
        self._layout.addSpacing(20)
        self._layout.addWidget(self._recommendation_section)
        self._layout.addStretch(0)
        self._reflow_action_buttons()

    def _calc_action_button_columns(self) -> int:
        margins = self._layout.contentsMargins()
        available = max(1, self.width() - margins.left() - margins.right())
        # cols ~= available width / button footprint, then clamp to [1, button_count].
        unit = ActionButtonMinWidth + ActionButtonSpacing
        cols = (available + ActionButtonSpacing) // unit
        return max(1, min(len(self._action_btns), cols))

    def _reflow_action_buttons(self):
        cols = self._calc_action_button_columns()
        # Avoid unnecessary teardown/rebuild when only height changed.
        if cols == self._action_cols:
            return
        self._action_cols = cols

        while self._action_layout.count():
            item = self._action_layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                self._action_layout.removeWidget(widget)

        for i, btn in enumerate(self._action_btns):
            row, col = divmod(i, cols)
            self._action_layout.addWidget(btn, row, col)

        for col in range(len(self._action_btns)):
            self._action_layout.setColumnStretch(col, 1 if col < cols else 0)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._reflow_action_buttons()

    def _clear_recommendation_panels(self):
        while self._recommendation_section_layout.count():
            item = self._recommendation_section_layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.setParent(None)
                widget.deleteLater()
        self._playlist_panel = None
        self._collection_panels = []

    def _create_panel_from_collection(self, provider, collection) -> Optional[Panel]:
        if not collection.models:
            return None
        if collection.type_ == CollectionType.only_playlists:
            return RecPlaylistsCollectionPanel(
                self._app,
                provider.identifier,
                collection,
                initial_row_count=1,
            )
        if collection.type_ == CollectionType.only_songs:
            return RecSongsCollectionPanel(
                self._app,
                provider.identifier,
                collection,
                initial_row_count=3,
            )
        if collection.type_ == CollectionType.only_albums:
            return RecAlbumsCollectionPanel(
                self._app,
                provider.identifier,
                collection,
                initial_row_count=2,
            )
        if collection.type_ == CollectionType.only_videos:
            return RecVideosCollectionPanel(
                self._app,
                provider.identifier,
                collection,
                initial_row_count=2,
            )
        logger.warning(
            "skip unsupported recommendation collection type: %s",
            collection.type_,
        )
        return None

    async def render(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        if pvd_ui is None:
            self._app.show_msg("wtf!")
            return

        self._clear_recommendation_panels()
        provider = pvd_ui.provider
        if isinstance(provider, SupportsRecListCollections):
            collections = await run_fn(
                lambda: provider.rec_list_collections(limit=RecCollectionsLimit)
            )
            for collection in collections:
                panel = self._create_panel_from_collection(provider, collection)
                if panel is None:
                    continue
                self._recommendation_section_layout.addWidget(panel)
                self._collection_panels.append(panel)
            if self._collection_panels:
                self._recommendation_section.show()
                for panel in self._collection_panels:
                    await panel.render()
            else:
                # Keep action buttons usable even when no collection can be shown.
                self._recommendation_section.hide()
        elif isinstance(provider, SupportsRecListDailyPlaylists):
            if self._playlist_panel is None:
                self._playlist_panel = RecPlaylistsPanel(
                    self._app,
                    provider,
                    initial_row_count=1,
                    # /rec is scoped to the current provider, so provider icon is
                    # redundant here.
                    show_icon=False,
                )
                self._playlist_panel.header.setText(t("music-customized-recommendation"))
                self._recommendation_section_layout.addWidget(self._playlist_panel)
            self._recommendation_section.show()
            await self._playlist_panel.render()
        else:
            # Keep action buttons usable even when provider has no daily playlists.
            self._recommendation_section.hide()

        self.daily_songs_btn.setEnabled(isinstance(provider, SupportsRecListDailySongs))
        self.rank_btn.setEnabled(isinstance(provider, SupportsToplist))
        self.heart_radar_btn.setEnabled(
            isinstance(provider, SupportsCurrentUserListRadioSongs)
        )
        self.dislike_btn.setEnabled(
            isinstance(provider, SupportsCurrentUserDislikeSongsReader)
        )

    def _on_heart_radar_clicked(self):
        pvd_ui = self._app.current_pvd_ui_mgr.get()
        assert pvd_ui is not None
        provider = pvd_ui.provider
        assert isinstance(provider, SupportsCurrentUserListRadioSongs)

        was_active = self._app.fm.is_active
        current_provider = self._heart_radar_provider if was_active else None
        fetch_func = provider.current_user_list_radio_songs
        if was_active and current_provider is provider:
            self._app.show_msg(t("music-radio-radar-activated"))
            return

        if was_active:
            self._app.fm.deactivate()

        self._app.fm.activate(fetch_func)
        self._heart_radar_provider = provider
        self._app.show_msg(
            t("music-radio-radar-changed")
            if was_active
            else t("music-radio-radar-activated")
        )



# flake8: noqa

import asyncio
import sys


def patch_qeventloop():
    # patch note::
    #
    # asyncio.wrap_future should set loop to self, otherwise, we can not call run_in_executor
    # in another thread(the loop not belongs to). The event loop implemented in asyncio
    # library also does this.

    from .compat import QEventLoop, QThreadExecutor

    def run_in_executor(self, executor, callback, *args):
        """Run callback in executor.

        If no executor is provided, the default executor will be used, which defers execution to
        a background thread.
        """
        # pylint: disable=all

        self._logger.debug('Running callback {} with args {} in executor'.format(callback, args))
        if isinstance(callback, asyncio.Handle):
            assert not args
            assert not isinstance(callback, asyncio.TimerHandle)
            if callback._cancelled:
                f = asyncio.Future()
                f.set_result(None)
                return f
            callback, args = callback.callback, callback.args

        if executor is None:
            self._logger.debug('Using default executor')
            executor = self.get_default_executor()

        if executor is None:
            self._logger.debug('Creating default executor')
            self.set_default_executor_v2(QThreadExecutor())
            executor = self.get_default_executor()

        return asyncio.wrap_future(executor.submit(callback, *args), loop=self)

    def get_default_executor(self):
        return getattr(self, 'default_executor', None)

    def set_default_executor_v2(self, executor):
        setattr(self, 'default_executor', executor)

        # QEventLoop will shutdown executor when loop is closed
        self.set_default_executor(executor)

    QEventLoop.get_default_executor = get_default_executor
    QEventLoop.set_default_executor_v2 = set_default_executor_v2
    QEventLoop.run_in_executor = run_in_executor


def patch_janus():
    """
    https://github.com/aio-libs/janus/issues/250
    """

    import janus

    janus.current_loop = asyncio.get_event_loop


def patch_mutagen():

    from mutagen.id3 import _specs

    old_decode_terminated = _specs.decode_terminated

    def decode_terminated(data, encoding, strict=True):
        """
        Some songs in China have their ID3 information encoded in GBK,
        but their metadata indicates latin1 encoding, so mutagen parsing
        returns incorrect results.

        Here, we did not choose to reimplement the original function,
        but process its return value, which better ensures our implementation
        does not affect other mutagen logic.

        However, this approach has an issue: the original function has already
        performed decode and encode, and we repeat encode and decode here,
        theoretically causing some performance overhead.

        We welcome interested peers to optimize.



        > **NOTE**: According to the ID3 2.3/2.4 specification, GBK encoding is *apparently*
        > illegal in the first place; the spec states only these four are supported:
        > latin1, utf16, utf_16_be, utf8.
        > Mutagen indeed implements according to the spec, but reality is always cruel.
        """
        orig_value, data = old_decode_terminated(data, encoding, strict)
        value = orig_value
        if encoding == 'latin1':
            value = orig_value.encode('latin1')
            # Refer to the deepin-music-player project’s easymp3.py for implementation
            for codec in ['gbk', 'big5', 'utf-8', 'iso-8559-15', 'cp1251']:
                try:
                    value = value.decode(codec)
                except (UnicodeDecodeError, LookupError):
                    # decode failed or codec not found
                    pass
                else:
                    break
            else:
                value = orig_value
        return value, data

    _specs.decode_terminated = decode_terminated


def patch_pydantic():
    """
    For some reason, some plugins already use pydantic v2 and they only use
    pydantic.v1 package. Make them compat with pydantic v1 in an hacker way :)
    """
    try:
        import pydantic.v1
    except ImportError:
        import pydantic
        sys.modules['pydantic.v1'] = pydantic



import asyncio

from qasync import QEventLoop, QThreadExecutor, AllEvents
from PyQt6.QtWidgets import QApplication


try:
    from qasync import DefaultQEventLoopPolicy
except ImportError:
    # From qasync>=0.28, the class DefaultQEventLoopPolicy is removed
    # when python version >= 3.12
    class DefaultQEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
        def new_event_loop(self):
            return QEventLoop(QApplication.instance() or QApplication(['FeelUOwn']))


# Work around Python 3.14 running-loop checks triggered by qasync shutdown.
class PatchedQEventLoop(QEventLoop):
    def run_until_complete(self, future):
        if self.is_running():
            raise RuntimeError("Event loop already running")

        future = asyncio.ensure_future(future, loop=self)

        def stop(*args):
            self.stop()

        future.add_done_callback(stop)
        try:
            self.run_forever()
        finally:
            future.remove_done_callback(stop)

        prev_loop = asyncio.events._get_running_loop()
        asyncio.events._set_running_loop(self)
        try:
            QApplication.instance().eventDispatcher().processEvents(AllEvents)
        finally:
            asyncio.events._set_running_loop(prev_loop)

        if not future.done():
            raise RuntimeError("Event loop stopped before Future completed.")

        return future.result()


__all__ = (
    'QEventLoop',
    'QThreadExecutor',
)



# -*- coding: utf-8 -*-
import os
import logging

from feeluown.i18n import t
from feeluown.utils import aio  # noqa
from .provider import provider  # noqa

DEFAULT_MUSIC_FOLDER = os.path.expanduser("~") + "/Music"
DEFAULT_MUSIC_EXTS = ["mp3", "ogg", "wma", "m4a", "m4v", "mp4", "flac", "ape", "wav"]

__alias__ = "本地音乐"
__feeluown_version__ = "1.1.0"
__version__ = "0.1a0"
__desc__ = "本地音乐"

logger = logging.getLogger(__name__)


def init_config(config):
    config.deffield(
        "MUSIC_FOLDERS",
        type_=list,
        default=[DEFAULT_MUSIC_FOLDER],
        desc="Supported list of music folders",
    )
    config.deffield(
        "MUSIC_FORMATS",
        type_=list,
        default=DEFAULT_MUSIC_EXTS,
        desc="Supported list of music formats",
    )
    config.deffield(
        "CORE_LANGUAGE", type_=str, default="auto", desc="Default display language"
    )
    config.deffield(
        "IDENTIFIER_DELIMITER",
        type_=str,
        default="",
        desc="Delimiter used when generating identifiers",
    )
    config.deffield(
        "EXPAND_ARTIST_SONGS",
        type_=bool,
        default=False,
        desc="Include songs from album artists' albums in that artist's song list",
    )
    config.deffield(
        "ARTIST_SPLITTER",
        type_=list,
        default=[",", "&"],
        desc="Delimiters for song artists",
    )
    config.deffield(
        "ARTIST_SPLITTER_IGNORANCE",
        type_=list,
        default=None,
        desc="Strings to be protected when using delimiters on artist info",
    )
    config.deffield(
        "SPLIT_ALBUM_ARTIST_NAME",
        type_=bool,
        default=False,
        desc="Allow splitting album artist names using delimiters",
    )


async def autoload(app):
    await aio.run_fn(provider.scan, app.config.local, app.config.local.MUSIC_FOLDERS)

    app.show_msg(t("local-tracks-scan-finished"))


def enable(app):
    logger.info("Register provider: %s", provider)
    app.library.register(provider)
    provider.initialize(app)

    app.started.connect(
        lambda *args: aio.create_task(autoload(*args)), weak=False, aioqueue=False
    )
    if app.mode & app.GuiMode:
        from .provider_ui import LocalProviderUi

        provider_ui = LocalProviderUi(app)
        app.pvd_ui_mgr.register(provider_ui)


def disable(app):
    logger.info("Oops, local provider cannot be disabled")



import sys

if sys.platform == "win32":
    # By default, it uses SimSun(宋体) on windows, which is a little ugly.
    # "Segoe UI Symbol" is used to render charactor symbols.
    # "Microsoft Yahei" is used to render chinese (and english).
    # Choose a default sans-serif font when the first two fonts do not work,
    FontFamilies = ["Segoe UI Symbol", "Microsoft YaHei", "sans-serif"]
elif sys.platform == "darwin":
    FontFamilies = ["arial", "Hiragino Sans GB", "sans-serif"]
else:
    FontFamilies = [
        "Helvetica Neue",
        "Helvetica",
        "Arial",
        "Microsoft Yahei",
        "Hiragino Sans GB",
        "Heiti SC",
        "WenQuanYi Micro Hei",
        "sans-serif",
    ]

# The width of ScrollBar on macOS is about 10.
ScrollBarWidth = 10



from .base import SerializerError, DeserializerError

# format Serializer mapping, like::
#
# {
#    'json':  JsonSerializer,
#    'plain': PlainSerializer
# }
_MAPPING = {}
_DE_MAPPING = {}


def register_serializer(type_, serializer_cls):
    _MAPPING[type_] = serializer_cls


def register_deserializer(type_, deserializer_cls):
    _DE_MAPPING[type_] = deserializer_cls


def get_serializer(format_):
    if not _MAPPING:
        register_serializer('plain', PlainSerializer)
        register_serializer('json', JsonSerializer)
        register_serializer('python', PythonSerializer)
    if format_ not in _MAPPING:
        raise SerializerError(f"Serializer for format:{format_} not found")
    return _MAPPING.get(format_)


def get_deserializer(format_: str):
    if not _DE_MAPPING:
        register_deserializer('python', PythonDeserializer)
    if format_ not in _DE_MAPPING:
        raise DeserializerError(f"Deserializer for format:{format_} not found")
    return _DE_MAPPING[format_]


def serialize(format_, obj, **options):
    """serialize python object defined in feeluown package

    :raises SerializerError:

    Usage::

        serialize('plain', song, as_line=True)
        serialize('plain', song, as_line=True, fetch=True)
        serialize('json', songs, indent=4)
        serialize('json', songs, indent=4, fetch=True)
        serialize('json', providers)
    """
    serializer = get_serializer(format_)(**options)
    return serializer.serialize(obj)


def deserialize(format_, obj, **options):
    deserializer = get_deserializer(format_)(**options)
    return deserializer.deserialize(obj)


from .base import SerializerMeta, SimpleSerializerMixin  # noqa
from .plain import PlainSerializer  # noqa
from .json_ import JsonSerializer  # noqa
from .python import PythonSerializer, PythonDeserializer  # noqa
from .objs import *  # noqa



from mutagen.mp3 import MP3
from mutagen.mp4 import MP4, AtomDataType
from mutagen.flac import FLAC


def read_audio_cover(fpath):
    """read audio cover binary data and format"""

    if fpath.endswith('mp3'):
        mp3 = MP3(fpath)
        apic = mp3.get('APIC:')
        if apic is not None:
            if apic.mime in ('image/jpg', 'image/jpeg'):
                fmt = 'jpg'
            else:
                fmt = 'png'
            return apic.data, fmt

    elif fpath.endswith('m4a'):
        mp4 = MP4(fpath)
        tags = mp4.tags
        if tags is not None:
            covers = tags.get('covr')
            if covers:
                cover = covers[0]
                if cover.imageformat == AtomDataType.JPEG:
                    fmt = 'jpg'
                else:
                    fmt = 'png'
                return cover, fmt

    elif fpath.endswith('flac'):
        flac = FLAC(fpath)
        covers = flac.pictures
        if covers:
            cover = covers[0]
            if cover.mime in ('image/jpg', 'image/jpeg'):
                fmt = 'jpg'
            else:
                fmt = 'png'
            return cover.data, fmt

    return None, None



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position
"""
TODO: The current logic in this module is very extensive,
including small parts such as music directory scanning and music library construction;
in theory, all these small parts can be extracted from it.
"""

import difflib
import logging
import re
import threading
from functools import wraps

from feeluown.excs import ProviderIOError

from feeluown.i18n import t
from feeluown.media import Media, Quality
from feeluown.library import AbstractProvider, ProviderV2, ModelType, SimpleSearchResult
from feeluown.utils.reader import create_reader
from feeluown.utils.utils import log_exectime
from feeluown.utils.audio import read_audio_cover
from .db import sort_album_func


logger = logging.getLogger(__name__)
SOURCE = 'local'


def wait_for_scan(func):
    """
    Some API of local provider can only return correct data after scan is finished.
    This decorator is designed to be used to decorate those API.
    """
    @wraps(func)
    def wrapper(this, *args, **kwargs):
        if not this._scan_finished.wait(timeout=1):
            raise ProviderIOError('scan is not still finished')
        return func(this, *args, **kwargs)
    return wrapper


class LocalProvider(AbstractProvider, ProviderV2):
    class meta:
        identifier = SOURCE
        name = t('local-tracks')

    def __init__(self):
        super().__init__()

        self._app = None
        self._scan_finished = threading.Event()

        from .db import DB

        self.db = DB('')

    def initialize(self, app):
        self._app = app

    def handle_with_path(self, path, **_):
        """
        handle ('/songs/{identifier}/cover/data')
        """
        p = re.compile(r'/songs/(\S+)/cover/data')
        m = p.match(path)
        if m is not None:
            try:
                identifier = m.group(1)
            except IndexError:
                return None
            else:
                fpath = self.db.get_song_fpath(identifier)
                if fpath:
                    return read_audio_cover(fpath)[0]
        return None

    @property
    def identifier(self):
        return SOURCE

    @property
    def name(self):
        return LocalProvider.meta.name

    def scan(self, config, paths, depth=3):
        exts = config.MUSIC_FORMATS
        self.db.scan(config, paths, depth, exts)
        self.db.after_scan()
        self._scan_finished.set()

    def use_model_v2(self, model_type):
        return model_type in (ModelType.song, ModelType.album, ModelType.artist)

    @wait_for_scan
    def song_get(self, identifier):
        """implements SupportsSongGet protocol."""
        return self.db.get_song(identifier)

    def song_list_quality(self, _):
        """implements SupportsSongMultiQuality protocol."""
        return [Quality.Audio.sq]

    @wait_for_scan
    def song_get_media(self, song, _):
        """implements SupportsSongMultiQuality protocol."""
        fpath = self.db.get_song_fpath(song.identifier)
        if fpath:
            return Media(fpath)

    @wait_for_scan
    def album_get(self, identifier):
        """Implement SupportsAlbumGet protocol."""
        return self.db.get_album(identifier)

    @wait_for_scan
    def artist_get(self, identifier):
        """Implement SupportsArtistGet protocol."""
        return self.db.get_artist(identifier)

    @wait_for_scan
    def artist_create_songs_rd(self, artist):
        """Implement SupportsArtistSongsReader protocol."""
        artist = self.model_get(artist.meta.model_type, artist.identifier)
        return create_reader(artist.hot_songs)

    @wait_for_scan
    def artist_create_albums_rd(self, artist):
        """Implement SupportsArtistAlbumsReader protocol."""
        albums = []
        for album in self.db.list_albums():
            for artist_ in album.artists:
                if artist_.identifier == artist.identifier:
                    albums.append(album)
                    continue
        albums.sort(key=sort_album_func, reverse=True)
        return create_reader(albums)

    @wait_for_scan
    def artist_create_contributed_albums_rd(self, artist):
        albums = self.db.list_albums_by_contributor(artist.identifier)
        albums.sort(key=sort_album_func, reverse=True)
        return create_reader(albums)

    @property
    def songs(self):
        return self.db.list_songs()

    @property
    def albums(self):
        return self.db.list_albums()

    @property
    def artists(self):
        return self.db.list_artists()

    @log_exectime
    @wait_for_scan
    def search(self, keyword, **kwargs):
        from .db import to_brief_song

        limit = kwargs.get('limit', 10)
        repr_song_map = dict()
        for song in self.songs:
            key = song.title + ' ' + song.artists_name
            repr_song_map[key] = song
        choices = repr_song_map.keys()
        if choices:
            # TODO: maybe use a more reasonable algorithm.
            result = difflib.get_close_matches(keyword, choices, n=limit, cutoff=0)
        else:
            result = []
        result_songs = [repr_song_map[each] for each in result]

        return SimpleSearchResult(
            q=keyword,
            songs=[to_brief_song(song) for song in result_songs]
        )


provider = LocalProvider()



# pylint: disable=import-error
from AppKit import NSImage, NSMakeRect, NSCompositingOperationSourceOver
from Foundation import NSMutableDictionary
from MediaPlayer import (
    MPMediaItemArtwork, MPNowPlayingInfoCenter, MPMediaItemPropertyArtwork,
)

from feeluown.utils import aio


class MacosMixin:

    def update_song_metadata(self, meta):
        super().update_song_metadata(meta)

        artwork = meta.get('artwork', '')
        artwork_uid = meta.get('uri', artwork)
        self._update_artwork(artwork, artwork_uid)

    def _update_artwork(self, artwork, artwork_uid):

        async def task():
            b = await self._app.img_mgr.get(artwork, artwork_uid)
            set_artwork(self.info_center, b)

        aio.run_afn(task)


def artwork_from_bytes(b: bytes) -> MPMediaItemArtwork:
    img = NSImage.alloc().initWithData_(b)

    def resize(size):
        new = NSImage.alloc().initWithSize_(size)
        new.lockFocus()
        img.drawInRect_fromRect_operation_fraction_(
            NSMakeRect(0, 0, size.width, size.height),
            NSMakeRect(0, 0, img.size().width, img.size().height),
            NSCompositingOperationSourceOver,
            1.0,
        )
        new.unlockFocus()
        return new

    art = MPMediaItemArtwork.alloc() \
        .initWithBoundsSize_requestHandler_(img.size(), resize)
    return art


def set_artwork(info_center: MPNowPlayingInfoCenter, b):
    current_nowplaying_info = info_center.nowPlayingInfo()
    if current_nowplaying_info is None:
        nowplaying_info = NSMutableDictionary.dictionary()
    else:
        nowplaying_info = current_nowplaying_info.mutableCopy()
    if b:
        nowplaying_info[MPMediaItemPropertyArtwork] = artwork_from_bytes(b)
    info_center.setNowPlayingInfo_(nowplaying_info)
    return True



from typing import TYPE_CHECKING, Sequence

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap, QGuiApplication
from PyQt6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel

from feeluown.i18n import t
from feeluown.library import (
    Collection,
    BriefAlbumModel,
    SupportsRecListDailyPlaylists,
    BriefVideoModel,
    BriefSongModel,
)
from feeluown.utils.reader import create_reader
from feeluown.utils.aio import run_fn
from feeluown.gui.widgets.header import LargeHeader
from feeluown.gui.widgets.img_card_list import (
    AlbumCardListView,
    AlbumCardListModel,
    AlbumFilterProxyModel,
    AlbumCardListDelegate,
    PlaylistCardListView,
    PlaylistCardListModel,
    PlaylistFilterProxyModel,
    PlaylistCardListDelegate,
    VideoCardListView,
    VideoCardListModel,
    VideoCardListDelegate,
)
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListView,
    SongMiniCardListDelegate,
    SongMiniCardListModel,
)
from feeluown.gui.widgets.selfpaint_btn import TriagleButton

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


PlaylistCardMinWidth = 140
PlaylistCardSpacing = 25
SongCardHeight = 50
SongCardPadding = (5, 5, 5, 5)
VideoCardMinWidth = 200


def create_song_list_view(app: "GuiApp", row_count: int) -> SongMiniCardListView:
    songs_list_view = SongMiniCardListView(
        no_scroll_v=True,
        fixed_row_count=row_count,
    )
    songs_list_view.setItemDelegate(
        SongMiniCardListDelegate(
            songs_list_view,
            card_height=SongCardHeight,
            card_padding=SongCardPadding,
        )
    )
    songs_list_view.play_song_needed.connect(app.playlist.play_model)
    return songs_list_view


def _resolve_collapsed_row_count(item_count: int, default_row_count: int) -> int:
    # Keep one-row collections compact; otherwise fixed-row layouts leave obvious
    # blank space in discovery panels.
    return 1 if item_count == 1 else default_row_count


class FoldButton(TriagleButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setToolTip(t("fold-tooltip"))
        self.setCheckable(True)
        self.toggled.connect(self.on_toggled)
        # Checked means folded, and show down direction. Click to unfold.
        self.setChecked(True)

    def on_toggled(self, checked):
        self.setToolTip(
            t("fold-expand") if checked else t("fold-collapse"),
        )
        self.set_direction("down" if checked else "up")


class Panel(QWidget):
    # Shared panel chrome for homepage/discovery recommendation blocks.
    _id_pixmap_cache = {}

    def __init__(self, title, body, pixmap, *, show_icon: bool = True):
        super().__init__(parent=None)

        self.icon_label = QLabel()
        self.header = LargeHeader(title)
        self.body = body

        self.icon_label.setFixedSize(20, 20)
        self.icon_label.setPixmap(pixmap)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(10)

        self._h_layout = QHBoxLayout()
        self._h_layout.setSpacing(5)
        self._layout.addLayout(self._h_layout)
        # Homepage may aggregate panels from different providers, while
        # provider-scoped pages can hide this icon for a cleaner header.
        if show_icon:
            self._h_layout.addWidget(self.icon_label)
        else:
            self.icon_label.hide()
        self._h_layout.addWidget(self.header)
        self._h_layout.addStretch(1)

        self.fold_unfold_btn = FoldButton(length=16)
        self._h_layout.addWidget(self.fold_unfold_btn)
        self._h_layout.addSpacing(20)

        self._layout.addWidget(self.body)

    @classmethod
    def get_provider_pixmap(cls, app: "GuiApp", provider_id, width=20):
        device_pixel_ratio = QGuiApplication.instance().devicePixelRatio()
        if provider_id in cls._id_pixmap_cache:
            return cls._id_pixmap_cache[provider_id]
        pvd_ui = app.pvd_ui_mgr.get(provider_id)
        if pvd_ui is None:
            svg = "icons:feeluown.png"
        else:
            svg = pvd_ui.get_colorful_svg()
        pixmap = QPixmap(svg).scaledToWidth(
            int(width * device_pixel_ratio), Qt.TransformationMode.SmoothTransformation
        )
        pixmap.setDevicePixelRatio(device_pixel_ratio)
        cls._id_pixmap_cache[provider_id] = pixmap
        return pixmap

    async def render(self):
        pass


class _PlaylistsPanelBase(Panel):
    def __init__(
        self,
        title: str,
        app: "GuiApp",
        provider_id: str,
        *,
        initial_row_count: int = 2,
    ):
        self._app = app
        self._initial_row_count = initial_row_count
        self._collapsed_row_count = initial_row_count
        self.playlist_list_view = PlaylistCardListView(
            no_scroll_v=True, fixed_row_count=self._initial_row_count
        )
        # Bind once at init to avoid duplicate signal connections after re-render.
        self.playlist_list_view.show_playlist_needed.connect(
            lambda model: self._app.browser.goto(model=model)
        )
        # Keep card visual style consistent between homepage and discovery page.
        self.playlist_list_view.setItemDelegate(
            PlaylistCardListDelegate(
                self.playlist_list_view,
                card_min_width=PlaylistCardMinWidth,
                card_spacing=PlaylistCardSpacing,
            )
        )
        pixmap = Panel.get_provider_pixmap(app, provider_id)
        super().__init__(
            title,
            self.playlist_list_view,
            pixmap,
            show_icon=True,
        )

        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)

    def _show_more_or_less(self, checked):
        # Folded keeps a predictable compact height; unfolded shows all rows.
        if checked:
            self.playlist_list_view.set_fixed_row_count(self._collapsed_row_count)
        else:
            self.playlist_list_view.set_fixed_row_count(-1)

    def set_playlists(self, playlists):
        if not playlists:
            return
        self._collapsed_row_count = _resolve_collapsed_row_count(
            len(playlists), self._initial_row_count
        )
        self.playlist_list_view.set_fixed_row_count(self._collapsed_row_count)
        model = PlaylistCardListModel.create(create_reader(playlists), self._app)
        filter_model = PlaylistFilterProxyModel()
        filter_model.setSourceModel(model)
        self.playlist_list_view.setModel(filter_model)


class RecPlaylistsPanel(_PlaylistsPanelBase):
    def __init__(
        self,
        app: "GuiApp",
        provider: SupportsRecListDailyPlaylists,
        *,
        initial_row_count: int = 2,
        show_icon: bool = True,
    ):
        self._provider = provider
        super().__init__(
            t("recommended-playlist"),
            app,
            provider.identifier,
            initial_row_count=initial_row_count,
        )
        if not show_icon:
            self.icon_label.hide()

    async def render(self):
        playlists = await run_fn(self._provider.rec_list_daily_playlists)
        self.set_playlists(playlists)


class RecPlaylistsCollectionPanel(_PlaylistsPanelBase):
    def __init__(
        self,
        app: "GuiApp",
        provider_id: str,
        collection: Collection,
        *,
        initial_row_count: int = 2,
    ):
        self._playlists = collection.models
        super().__init__(
            collection.name,
            app,
            provider_id,
            initial_row_count=initial_row_count,
        )
        self.icon_label.hide()

    async def render(self):
        self.set_playlists(self._playlists)


class RecSongsCollectionPanel(Panel):
    def __init__(
        self,
        app: "GuiApp",
        provider_id: str,
        collection: Collection,
        *,
        initial_row_count: int = 3,
    ):
        self._app = app
        self._songs: Sequence[BriefSongModel] = collection.models
        self._initial_row_count = initial_row_count
        self._collapsed_row_count = initial_row_count
        self.songs_list_view = create_song_list_view(
            self._app, self._initial_row_count
        )

        pixmap = Panel.get_provider_pixmap(app, provider_id)
        super().__init__(collection.name, self.songs_list_view, pixmap, show_icon=False)
        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)

    def _show_more_or_less(self, checked):
        if checked:
            self.songs_list_view.set_fixed_row_count(self._collapsed_row_count)
        else:
            self.songs_list_view.set_fixed_row_count(-1)

    async def render(self):
        if not self._songs:
            return
        self._collapsed_row_count = _resolve_collapsed_row_count(
            len(self._songs), self._initial_row_count
        )
        self.songs_list_view.set_fixed_row_count(self._collapsed_row_count)
        model = SongMiniCardListModel.create(create_reader(self._songs), self._app)
        self.songs_list_view.setModel(model)


class RecAlbumsCollectionPanel(Panel):
    def __init__(
        self,
        app: "GuiApp",
        provider_id: str,
        collection: Collection,
        *,
        initial_row_count: int = 2,
    ):
        self._app = app
        self._albums: Sequence[BriefAlbumModel] = collection.models
        self._initial_row_count = initial_row_count
        self._collapsed_row_count = initial_row_count
        self.album_list_view = AlbumCardListView(
            no_scroll_v=True, fixed_row_count=self._initial_row_count
        )
        self.album_list_view.show_album_needed.connect(
            lambda model: self._app.browser.goto(model=model)
        )
        self.album_list_view.setItemDelegate(
            AlbumCardListDelegate(
                self.album_list_view,
                card_min_width=PlaylistCardMinWidth,
                card_spacing=PlaylistCardSpacing,
            )
        )

        pixmap = Panel.get_provider_pixmap(app, provider_id)
        super().__init__(
            collection.name, self.album_list_view, pixmap, show_icon=False
        )
        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)

    def _show_more_or_less(self, checked):
        if checked:
            self.album_list_view.set_fixed_row_count(self._collapsed_row_count)
        else:
            self.album_list_view.set_fixed_row_count(-1)

    async def render(self):
        if not self._albums:
            return
        self._collapsed_row_count = _resolve_collapsed_row_count(
            len(self._albums), self._initial_row_count
        )
        self.album_list_view.set_fixed_row_count(self._collapsed_row_count)
        model = AlbumCardListModel.create(create_reader(self._albums), self._app)
        filter_model = AlbumFilterProxyModel()
        filter_model.setSourceModel(model)
        self.album_list_view.setModel(filter_model)


class RecVideosCollectionPanel(Panel):
    def __init__(
        self,
        app: "GuiApp",
        provider_id: str,
        collection: Collection,
        *,
        initial_row_count: int = 2,
    ):
        self._app = app
        self._videos: Sequence[BriefVideoModel] = collection.models
        self._initial_row_count = initial_row_count
        self._collapsed_row_count = initial_row_count
        self.video_list_view = VideoCardListView(
            no_scroll_v=True, fixed_row_count=self._initial_row_count
        )
        self.video_list_view.setItemDelegate(
            VideoCardListDelegate(
                self.video_list_view,
                card_min_width=VideoCardMinWidth,
            )
        )
        self.video_list_view.play_video_needed.connect(self._app.playlist.play_model)

        pixmap = Panel.get_provider_pixmap(app, provider_id)
        super().__init__(
            collection.name, self.video_list_view, pixmap, show_icon=False
        )
        self.fold_unfold_btn.clicked.connect(self._show_more_or_less)

    def _show_more_or_less(self, checked):
        if checked:
            self.video_list_view.set_fixed_row_count(self._collapsed_row_count)
        else:
            self.video_list_view.set_fixed_row_count(-1)

    async def render(self):
        if not self._videos:
            return
        self._collapsed_row_count = _resolve_collapsed_row_count(
            len(self._videos), self._initial_row_count
        )
        self.video_list_view.set_fixed_row_count(self._collapsed_row_count)
        model = VideoCardListModel.create(self._videos, self._app)
        self.video_list_view.setModel(model)



from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QModelIndex, QItemSelectionModel
from PyQt6.QtWidgets import QMenu, QAbstractItemView

from feeluown.utils.aio import run_fn, run_afn
from feeluown.library import (
    BriefSongModel,
    SupportsCurrentUserDislikeAddSong,
    SupportsCurrentUser,
)
from feeluown.player import PlaylistMode
from feeluown.gui.components import SongMenuInitializer
from feeluown.gui.helpers import fetch_cover_wrapper
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListView,
    SongMiniCardListModel,
)
from feeluown.i18n import t
from feeluown.utils.reader import create_reader


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class PlayerPlaylistModel(SongMiniCardListModel):
    """
    this is a singleton class (ensured by PlayerPlaylistView)
    """

    def __init__(self, playlist, *args, **kwargs):
        reader = create_reader(playlist.list())
        super().__init__(reader, *args, **kwargs)

        self._playlist = playlist
        self._playlist.songs_added.connect(self.on_songs_added)
        self._playlist.songs_removed.connect(self.on_songs_removed)
        self._playlist.songs_reordered.connect(self.on_songs_reordered)

    def flags(self, index):
        flags = super().flags(index)
        song = index.data(Qt.ItemDataRole.UserRole)[0]
        if self._playlist.is_bad(song):
            # Disable bad song.
            flags &= ~Qt.ItemFlag.ItemIsEnabled
        return flags

    def on_songs_added(self, index, count):
        self.beginInsertRows(QModelIndex(), index, index + count - 1)
        # Insert from tail to front.
        while count > 0:
            self._items.insert(index, self._playlist[index + count - 1])
            count -= 1
        self.endInsertRows()

    def on_songs_removed(self, index, count):
        self.beginRemoveRows(QModelIndex(), index, index + count - 1)
        while count > 0:
            self._items.pop(index + count - 1)
            count -= 1
        self.endRemoveRows()

    def on_songs_reordered(self, index, count):
        self.on_songs_removed(index, count)
        self.on_songs_added(index, count)


class PlayerPlaylistView(SongMiniCardListView):
    _model = None

    def __init__(self, app: "GuiApp", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

        self.play_song_needed.connect(self._app.playlist.play_model)
        if PlayerPlaylistView._model is None:
            PlayerPlaylistView._model = PlayerPlaylistModel(
                self._app.playlist,
                fetch_cover_wrapper(self._app),
            )
        self.setModel(PlayerPlaylistView._model)

    def contextMenuEvent(self, e):
        indexes = self.selectedIndexes()
        if not indexes:
            return

        songs = [index.data(Qt.ItemDataRole.UserRole)[0] for index in indexes]
        menu = QMenu()
        if self._app.playlist.mode is PlaylistMode.fm:
            btn_text = t("fm-radio-current-song-dislike")
        else:
            btn_text = t("track-playlist-remove")
        action = menu.addAction(btn_text)
        action.triggered.connect(lambda: self._remove_songs(songs))

        # Just hide the action instead of making it disabled when conditions are not met.
        # Because this function is not supported by most providers,
        # and (I think) most users do not use it frequently.
        if len(songs) == 1:
            song = songs[0]
            provider = self._app.library.get(song.source)
            if (
                isinstance(provider, SupportsCurrentUserDislikeAddSong)
                and isinstance(provider, SupportsCurrentUser)
                and provider.has_current_user()
            ):
                action_dislike = menu.addAction(t("track-provider-blacklist-add"))
                action_dislike.triggered.connect(
                    lambda: run_afn(self._dislike_and_remove_songs, songs)
                )

        if len(songs) == 1:
            menu.addSeparator()
            SongMenuInitializer(self._app, songs[0]).apply(menu)
        menu.exec(e.globalPos())

    def scroll_to_current_song(self):
        """Scroll to the current song, and select it to highlight it."""
        current_song = self._app.playlist.current_song
        songs = self._app.playlist.list()
        if current_song is not None:
            model = self.model()
            row = songs.index(current_song)
            index = model.index(row, 0)
            # In order to highlight the current song.
            self.selectionModel().select(
                index, QItemSelectionModel.SelectionFlag.SelectCurrent
            )
            self.scrollTo(index, QAbstractItemView.ScrollHint.PositionAtCenter)

    async def _dislike_and_remove_songs(self, songs):
        song: BriefSongModel = songs[0]
        provider = self._app.library.get(song.source)
        assert isinstance(provider, SupportsCurrentUserDislikeAddSong)
        self._app.show_msg(t("track-provider-blacklist-adding"), timeout=3000)
        ok = await run_fn(provider.current_user_dislike_add_song, song)
        if ok:
            self._app.show_msg(t("track-provider-blacklist-add-succ"))
        else:
            self._app.show_msg(t("track-provider-blacklist-add-fail"), timeout=3000)
        self._remove_songs(songs)

    def _remove_songs(self, songs):
        for song in songs:
            playlist_songs = self._app.playlist.list()
            if (
                self._app.playlist.mode is PlaylistMode.fm
                # playlist_songs should not be empty, just for robustness
                and playlist_songs
                and song == self._app.playlist.current_song
                and playlist_songs[-1] == song
            ):
                self._app.show_msg(
                    t("track-radio-mode-remove-latest"),
                    timeout=3000,
                )
                self._app.playlist.next()
            else:
                self._app.playlist.remove(song)



import asyncio
import logging
from functools import partial
from packaging import version

import requests
from requests.exceptions import ConnectionError, Timeout

from feeluown import __version__
from feeluown.i18n import t

logger = logging.getLogger(__name__)


class VersionManager(object):
    def __init__(self, app):
        self._app = app

        self._app.started.connect(self.on_app_started)

    def on_app_started(self, *args):
        loop = asyncio.get_running_loop()
        loop.call_later(10, partial(loop.create_task, self.check_release()))

    async def check_release(self):
        loop = asyncio.get_event_loop()

        logger.info("Checking for updates...")
        try:
            resp = await loop.run_in_executor(
                None,
                partial(requests.get, "https://pypi.org/pypi/feeluown/json", timeout=2),
            )
        except (ConnectionError, Timeout) as e:
            logger.warning(e)
            logger.warning("Update check failed!")
        else:
            rv = resp.json()
            latest = version.parse(rv["info"]["version"])
            current = version.parse(__version__)

            if latest > current:
                msg_en = t(
                    "new-version-found",
                    locale="en-US",
                    latestVer=latest,
                    currentVer=current,
                )
                msg = t("new-version-found", latestVer=latest, currentVer=current)
                logger.warning(msg_en)
                if self._app.mode & self._app.GuiMode:
                    self._app.show_msg(msg)
            else:
                msg_en = t("already-updated", locale="en-US", latestVer=latest)
                msg = t("already-updated", latestVer=latest)
                logger.info(msg_en)
                if self._app.mode & self._app.GuiMode:
                    self._app.show_msg(msg)


if __name__ == "__main__":
    """
    Test the basic behavior of VersionManager
    """

    import logging

    logging.basicConfig()
    logger.setLevel(logging.DEBUG)

    class App(object):
        GuiMode = 0x10
        mode = 0x01

    app = App()
    mgr = VersionManager(app)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(loop.create_task(mgr.check_release()))



from typing import Any

from mcp.server.fastmcp import FastMCP
from feeluown.app import App, get_app
from feeluown.library import (
    ResolveFailed,
    ResolverNotFound,
    resolve,
    reverse,
    SearchType,
    BriefAlbumModel,
    BriefArtistModel,
    BriefPlaylistModel,
    BriefSongModel,
    BriefVideoModel,
)
from feeluown.library.provider_protocol import (
    SupportsAlbumGet,
    SupportsAlbumSongsReader,
    SupportsArtistGet,
    SupportsArtistAlbumsReader,
    SupportsArtistContributedAlbumsReader,
    SupportsArtistSongsReader,
    SupportsCurrentUser,
    SupportsCurrentUserFavAlbumsReader,
    SupportsCurrentUserFavArtistsReader,
    SupportsCurrentUserFavPlaylistsReader,
    SupportsCurrentUserFavSongsReader,
    SupportsCurrentUserFavVideosReader,
    SupportsCurrentUserListPlaylists,
    SupportsCurrentUserListRadioSongs,
    SupportsPlaylistSongsReader,
    SupportsRecACollectionOfSongs,
    SupportsRecACollectionOfVideos,
    SupportsRecListCollections,
    SupportsRecListDailyAlbums,
    SupportsRecListDailyPlaylists,
    SupportsRecListDailySongs,
    SupportsPlaylistGet,
    SupportsSongSimilar,
    SupportsToplist,
    SupportsPlaylistCreateByName,
    SupportsPlaylistDelete,
    SupportsPlaylistAddSong,
    SupportsPlaylistRemoveSong,
    SupportsSongGet,
    SupportsSongHotComments,
    SupportsSongMV,
    SupportsSongLyric,
    SupportsSongWebUrl,
    SupportsVideoGet,
    SupportsVideoWebUrl,
)
from feeluown.serializers import serialize


mcp = FastMCP("FeelUOwn")
_PROTOCOLS = (
    SupportsAlbumGet,
    SupportsAlbumSongsReader,
    SupportsArtistGet,
    SupportsArtistAlbumsReader,
    SupportsArtistContributedAlbumsReader,
    SupportsArtistSongsReader,
    SupportsCurrentUser,
    SupportsCurrentUserFavAlbumsReader,
    SupportsCurrentUserFavArtistsReader,
    SupportsCurrentUserFavPlaylistsReader,
    SupportsCurrentUserFavSongsReader,
    SupportsCurrentUserFavVideosReader,
    SupportsCurrentUserListPlaylists,
    SupportsCurrentUserListRadioSongs,
    SupportsPlaylistSongsReader,
    SupportsRecACollectionOfSongs,
    SupportsRecACollectionOfVideos,
    SupportsRecListCollections,
    SupportsRecListDailyAlbums,
    SupportsRecListDailyPlaylists,
    SupportsRecListDailySongs,
    SupportsPlaylistGet,
    SupportsSongSimilar,
    SupportsToplist,
    SupportsPlaylistCreateByName,
    SupportsPlaylistDelete,
    SupportsPlaylistAddSong,
    SupportsPlaylistRemoveSong,
    SupportsSongGet,
    SupportsSongHotComments,
    SupportsSongMV,
    SupportsSongLyric,
    SupportsSongWebUrl,
    SupportsVideoGet,
    SupportsVideoWebUrl,
)


def _require_app() -> App:
    app = get_app()
    if app is None:
        raise RuntimeError("app is not initialized")
    return app


def _player_nowplaying_metadata() -> dict[str, Any] | None:
    app = _require_app()
    return serialize("python", app.player.current_metadata)


def _playlist_list() -> list[dict[str, Any]]:
    app = _require_app()
    return serialize("python", app.playlist.list())


def _player_play_media_by_uri(uri: str) -> bool:
    app = _require_app()
    for model in app.playlist.list():
        if reverse(model) == uri:
            app.playlist.play_model(model)
            return True
    return False


def _current_song_uri() -> str | None:
    app = _require_app()
    song = app.playlist.current_song
    if song is None:
        return None
    return reverse(song)


def _provider_protocols(provider) -> list[str]:
    return [proto.__name__ for proto in _PROTOCOLS if isinstance(provider, proto)]


def _provider_from_id(provider_id: str):
    app = _require_app()
    return app.library.get(provider_id)


def _limit_search_payload(payload: dict[str, Any], limit: int | None) -> dict[str, Any]:
    if limit is None:
        return payload
    limited: dict[str, Any] = {}
    for key, value in payload.items():
        if isinstance(value, list):
            limited[key] = value[: max(limit, 0)]
        else:
            limited[key] = value
    return limited


def _limit_models(models: list[Any], limit: int | None) -> list[Any]:
    if limit is None:
        return list(models)
    return list(models)[: max(limit, 0)]


def _serialize_collection(collection) -> dict[str, Any]:
    return {
        "name": collection.name,
        "type": collection.type_.name,
        "description": collection.description,
        "models": serialize("python", collection.models),
    }


def _provider_model_get(
    provider_id: str, identifier: str, protocol, getter_name: str
) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, protocol):
        return None
    try:
        model = getattr(provider, getter_name)(identifier)
    except Exception:
        return None
    if model is None:
        return None
    return serialize("python", model)


def _provider_model_list(
    provider_id: str,
    identifier: str,
    protocol,
    list_method_name: str,
    model_builder,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, protocol):
        return None
    try:
        model = model_builder(provider_id, identifier)
        reader = getattr(provider, list_method_name)(model)
        models = reader.readall() if hasattr(reader, "readall") else list(reader)
    except Exception:
        return None
    return serialize("python", _limit_models(models, limit))


def _provider_noarg_list(
    provider_id: str,
    protocol,
    list_method_name: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, protocol):
        return None
    try:
        reader = getattr(provider, list_method_name)()
        models = reader.readall() if hasattr(reader, "readall") else list(reader)
    except Exception:
        return None
    return serialize("python", _limit_models(models, limit))


def _build_brief_playlist(provider_id: str, playlist_id: str) -> BriefPlaylistModel:
    return BriefPlaylistModel(identifier=playlist_id, source=provider_id)


def _build_brief_album(provider_id: str, album_id: str) -> BriefAlbumModel:
    return BriefAlbumModel(identifier=album_id, source=provider_id)


def _build_brief_artist(provider_id: str, artist_id: str) -> BriefArtistModel:
    return BriefArtistModel(identifier=artist_id, source=provider_id)


def _build_brief_song(provider_id: str, song_id: str) -> BriefSongModel:
    return BriefSongModel(identifier=song_id, source=provider_id)


def _build_brief_video(provider_id: str, video_id: str) -> BriefVideoModel:
    return BriefVideoModel(identifier=video_id, source=provider_id)


@mcp.tool()
def player_nowplaying_metadata() -> dict[str, Any] | None:
    """
    Get the metadata of the currently playing track.
    """
    return _player_nowplaying_metadata()


@mcp.resource("fuo://player/playlist")
def playlist_list() -> list[dict[str, Any]]:
    """
    List all tracks in the current playlist queue.
    Each item includes a `uri` field.
    """
    return _playlist_list()


@mcp.resource("fuo://player/nowplaying")
def nowplaying() -> dict[str, Any] | None:
    """
    Get the current playing track info.
    """
    metadata = _player_nowplaying_metadata()
    if metadata is None:
        return None
    return {
        "uri": _current_song_uri(),
        "metadata": metadata,
    }


@mcp.resource("fuo://library/providers")
def library_providers() -> list[dict[str, Any]]:
    app = _require_app()
    providers = []
    for provider in app.library.list():
        providers.append(
            {
                "id": provider.identifier,
                "name": provider.name,
            }
        )
    return providers


@mcp.tool()
def player_play_media_by_uri(uri: str) -> bool:
    """
    Play a track by URI if it exists in the current playlist queue.
    """
    return _player_play_media_by_uri(uri)


@mcp.tool()
def player_toggle() -> None:
    _require_app().player.toggle()


@mcp.tool()
def player_play() -> None:
    _require_app().player.resume()


@mcp.tool()
def player_pause() -> None:
    _require_app().player.pause()


@mcp.tool()
def player_seek(position: float) -> None:
    _require_app().player.position = position


@mcp.tool()
def player_status() -> dict[str, Any]:
    app = _require_app()
    player = app.player
    playlist = app.playlist
    return {
        "state": player.state.name,
        "position": player.position,
        "duration": player.duration,
        "volume": player.volume,
        "playback_mode": playlist.playback_mode.name,
        "nowplaying": nowplaying(),
    }


@mcp.tool()
def playlist_next() -> None:
    _require_app().playlist.next()


@mcp.tool()
def playlist_previous() -> None:
    _require_app().playlist.previous()


@mcp.tool()
def playlist_clear() -> None:
    _require_app().playlist.clear()


@mcp.tool()
def playlist_add_uri(uri: str) -> bool:
    app = _require_app()
    try:
        model = resolve(uri)
    except (ResolveFailed, ResolverNotFound):
        return False
    app.playlist.add(model)
    return True


@mcp.tool()
def playlist_play_uri(uri: str) -> bool:
    app = _require_app()
    try:
        model = resolve(uri)
    except (ResolveFailed, ResolverNotFound):
        return False
    app.playlist.add(model)
    app.playlist.play_model(model)
    return True


@mcp.tool()
def provider_capabilities(provider_id: str) -> dict[str, Any] | None:
    app = _require_app()
    provider = app.library.get(provider_id)
    if provider is None:
        return None
    return {
        "id": provider.identifier,
        "name": provider.name,
        "protocols": _provider_protocols(provider),
    }


@mcp.tool()
def provider_search(
    provider_id: str,
    keyword: str,
    type_in: str | None = None,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    Search provider resources. Returns a list of results per search type.
    """
    provider = _provider_from_id(provider_id)
    if provider is None:
        return None
    try:
        types = SearchType.batch_parse(type_in) if type_in else [SearchType.so]
    except ValueError:
        return None
    results = []
    for type_ in types:
        try:
            result = provider.search(keyword=keyword, type_=type_)
        except Exception:
            continue
        if result is None:
            continue
        payload = serialize("python", result)
        payload = _limit_search_payload(payload, limit)
        results.append(
            {
                "type": type_.value,
                "source": provider.identifier,
                "result": payload,
            }
        )
    return results


@mcp.tool()
def provider_rec_list_daily_songs(
    provider_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecListDailySongs):
        return None
    try:
        songs = provider.rec_list_daily_songs()
    except Exception:
        return None
    if songs is None:
        return None
    return serialize("python", _limit_models(songs, limit))


@mcp.tool()
def provider_rec_list_daily_playlists(
    provider_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecListDailyPlaylists):
        return None
    try:
        playlists = provider.rec_list_daily_playlists()
    except Exception:
        return None
    if playlists is None:
        return None
    return serialize("python", _limit_models(playlists, limit))


@mcp.tool()
def provider_rec_list_daily_albums(
    provider_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecListDailyAlbums):
        return None
    try:
        albums = provider.rec_list_daily_albums()
    except Exception:
        return None
    if albums is None:
        return None
    return serialize("python", _limit_models(albums, limit))


@mcp.tool()
def provider_rec_list_collections(
    provider_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecListCollections):
        return None
    try:
        collections = provider.rec_list_collections(limit=limit)
    except Exception:
        return None
    if collections is None:
        return None
    return [_serialize_collection(collection) for collection in collections]


@mcp.tool()
def provider_rec_a_collection_of_songs(provider_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecACollectionOfSongs):
        return None
    try:
        collection = provider.rec_a_collection_of_songs()
    except Exception:
        return None
    return _serialize_collection(collection)


@mcp.tool()
def provider_rec_a_collection_of_videos(provider_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsRecACollectionOfVideos):
        return None
    try:
        collection = provider.rec_a_collection_of_videos()
    except Exception:
        return None
    return _serialize_collection(collection)


@mcp.tool()
def provider_toplist_list(
    provider_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsToplist):
        return None
    try:
        toplists = provider.toplist_list()
    except Exception:
        return None
    if toplists is None:
        return None
    return serialize("python", _limit_models(toplists, limit))


@mcp.tool()
def provider_toplist_get(provider_id: str, toplist_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsToplist):
        return None
    try:
        playlist = provider.toplist_get(toplist_id)
    except Exception:
        return None
    if playlist is None:
        return None
    return serialize("python", playlist)


@mcp.tool()
def provider_song_get_lyric(provider_id: str, song_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsSongLyric):
        return None
    try:
        lyric = provider.song_get_lyric(_build_brief_song(provider_id, song_id))
    except Exception:
        return None
    if lyric is None:
        return None
    return serialize("python", lyric)


@mcp.tool()
def provider_song_get_web_url(provider_id: str, song_id: str) -> str | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsSongWebUrl):
        return None
    try:
        return provider.song_get_web_url(_build_brief_song(provider_id, song_id))
    except Exception:
        return None


@mcp.tool()
def provider_song_get_mv(provider_id: str, song_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsSongMV):
        return None
    try:
        video = provider.song_get_mv(_build_brief_song(provider_id, song_id))
    except Exception:
        return None
    if video is None:
        return None
    return serialize("python", video)


@mcp.tool()
def provider_song_list_similar(
    provider_id: str, song_id: str, limit: int | None = None
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsSongSimilar):
        return None
    try:
        songs = provider.song_list_similar(_build_brief_song(provider_id, song_id))
    except Exception:
        return None
    if songs is None:
        return None
    return serialize("python", _limit_models(songs, limit))


@mcp.tool()
def provider_song_list_hot_comments(
    provider_id: str,
    song_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsSongHotComments):
        return None
    try:
        comments = provider.song_list_hot_comments(
            _build_brief_song(provider_id, song_id)
        )
    except Exception:
        return None
    if comments is None:
        return None
    return serialize("python", _limit_models(comments, limit))


@mcp.tool()
def provider_video_get_web_url(provider_id: str, video_id: str) -> str | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsVideoWebUrl):
        return None
    try:
        return provider.video_get_web_url(_build_brief_video(provider_id, video_id))
    except Exception:
        return None


@mcp.tool()
def provider_playlist_create_by_name(
    provider_id: str,
    name: str,
) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsPlaylistCreateByName):
        return None
    try:
        playlist = provider.playlist_create_by_name(name)
    except Exception:
        return None
    if playlist is None:
        return None
    return serialize("python", playlist)


@mcp.tool()
def provider_playlist_delete(provider_id: str, playlist_id: str) -> bool | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsPlaylistDelete):
        return None
    try:
        return bool(provider.playlist_delete(playlist_id))
    except Exception:
        return None


@mcp.tool()
def provider_playlist_add_song(
    provider_id: str,
    playlist_id: str,
    song_id: str,
) -> bool | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsPlaylistAddSong):
        return None
    try:
        return bool(
            provider.playlist_add_song(
                _build_brief_playlist(provider_id, playlist_id),
                _build_brief_song(provider_id, song_id),
            )
        )
    except Exception:
        return None


@mcp.tool()
def provider_playlist_remove_song(
    provider_id: str,
    playlist_id: str,
    song_id: str,
) -> bool | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsPlaylistRemoveSong):
        return None
    try:
        return bool(
            provider.playlist_remove_song(
                _build_brief_playlist(provider_id, playlist_id),
                _build_brief_song(provider_id, song_id),
            )
        )
    except Exception:
        return None


@mcp.tool()
def provider_current_user_get(provider_id: str) -> dict[str, Any] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsCurrentUser):
        return None
    try:
        if hasattr(provider, "get_current_user_or_none"):
            user = provider.get_current_user_or_none()
        else:
            user = provider.get_current_user()
    except Exception:
        return None
    if user is None:
        return None
    return serialize("python", user)


@mcp.tool()
def provider_current_user_list_playlists(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserListPlaylists,
        "current_user_list_playlists",
        limit,
    )


@mcp.tool()
def provider_current_user_list_radio_songs(
    provider_id: str,
    count: int = 20,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    provider = _provider_from_id(provider_id)
    if provider is None or not isinstance(provider, SupportsCurrentUserListRadioSongs):
        return None
    try:
        songs = provider.current_user_list_radio_songs(count)
    except Exception:
        return None
    if songs is None:
        return None
    return serialize("python", _limit_models(songs, limit))


@mcp.tool()
def provider_current_user_fav_list_songs(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserFavSongsReader,
        "current_user_fav_create_songs_rd",
        limit,
    )


@mcp.tool()
def provider_current_user_fav_list_albums(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserFavAlbumsReader,
        "current_user_fav_create_albums_rd",
        limit,
    )


@mcp.tool()
def provider_current_user_fav_list_artists(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserFavArtistsReader,
        "current_user_fav_create_artists_rd",
        limit,
    )


@mcp.tool()
def provider_current_user_fav_list_playlists(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserFavPlaylistsReader,
        "current_user_fav_create_playlists_rd",
        limit,
    )


@mcp.tool()
def provider_current_user_fav_list_videos(
    provider_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    return _provider_noarg_list(
        provider_id,
        SupportsCurrentUserFavVideosReader,
        "current_user_fav_create_videos_rd",
        limit,
    )


@mcp.tool()
def provider_playlist_list_songs(
    provider_id: str,
    playlist_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    List songs in a provider playlist.
    """
    return _provider_model_list(
        provider_id,
        playlist_id,
        SupportsPlaylistSongsReader,
        "playlist_create_songs_rd",
        _build_brief_playlist,
        limit,
    )


@mcp.tool()
def provider_album_list_songs(
    provider_id: str,
    album_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    List songs in a provider album.
    """
    return _provider_model_list(
        provider_id,
        album_id,
        SupportsAlbumSongsReader,
        "album_create_songs_rd",
        _build_brief_album,
        limit,
    )


@mcp.tool()
def provider_artist_list_songs(
    provider_id: str,
    artist_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    List songs for a provider artist.
    """
    return _provider_model_list(
        provider_id,
        artist_id,
        SupportsArtistSongsReader,
        "artist_create_songs_rd",
        _build_brief_artist,
        limit,
    )


@mcp.tool()
def provider_artist_list_albums(
    provider_id: str,
    artist_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    List albums for a provider artist.
    """
    return _provider_model_list(
        provider_id,
        artist_id,
        SupportsArtistAlbumsReader,
        "artist_create_albums_rd",
        _build_brief_artist,
        limit,
    )


@mcp.tool()
def provider_artist_list_contributed_albums(
    provider_id: str,
    artist_id: str,
    limit: int | None = None,
) -> list[dict[str, Any]] | None:
    """
    List contributed albums for a provider artist.
    """
    return _provider_model_list(
        provider_id,
        artist_id,
        SupportsArtistContributedAlbumsReader,
        "artist_create_contributed_albums_rd",
        _build_brief_artist,
        limit,
    )


@mcp.tool()
def provider_song_get(provider_id: str, song_id: str) -> dict[str, Any] | None:
    """
    Get a song model by provider and identifier.
    """
    return _provider_model_get(provider_id, song_id, SupportsSongGet, "song_get")


@mcp.tool()
def provider_album_get(provider_id: str, album_id: str) -> dict[str, Any] | None:
    """
    Get an album model by provider and identifier.
    """
    return _provider_model_get(provider_id, album_id, SupportsAlbumGet, "album_get")


@mcp.tool()
def provider_artist_get(provider_id: str, artist_id: str) -> dict[str, Any] | None:
    """
    Get an artist model by provider and identifier.
    """
    return _provider_model_get(provider_id, artist_id, SupportsArtistGet, "artist_get")


@mcp.tool()
def provider_playlist_get(provider_id: str, playlist_id: str) -> dict[str, Any] | None:
    """
    Get a playlist model by provider and identifier.
    """
    return _provider_model_get(
        provider_id, playlist_id, SupportsPlaylistGet, "playlist_get"
    )


@mcp.tool()
def provider_video_get(provider_id: str, video_id: str) -> dict[str, Any] | None:
    """
    Get a video model by provider and identifier.
    """
    return _provider_model_get(provider_id, video_id, SupportsVideoGet, "video_get")


def run_mcp_server(host: str = "127.0.0.1", port: int = 23335, debug: bool = False):
    """
    Run the MCP server in Streamable HTTP mode.
    """
    mcp.settings.host = host
    mcp.settings.port = port
    mcp.settings.debug = debug
    mcp.settings.log_level = "DEBUG" if debug else "WARNING"
    return mcp.run_streamable_http_async()



from enum import Enum
from dataclasses import dataclass
from typing import List

from .models import BaseModel


class CollectionType(Enum):
    """Collection type enumeration"""

    # These two values are only used in local collection.
    sys_library = 16
    sys_pool = 13

    only_songs = 1
    only_artists = 2
    only_albums = 3
    only_playlists = 4
    only_lyrics = 5
    only_videos = 6

    only_users = 17
    only_comments = 18

    mixed = 32


@dataclass
class Collection:
    """
    Differences between a collection and a playlist
    - A collection has no identifier in general.
    - A collection may have songs, albums and artists.
    """
    name: str
    type_: CollectionType
    models: List[BaseModel]
    description: str = ''



"""
ImgList model and delegate

- ImgCardListDelegate
- ImgCardListModel

By default, we think the proper width of a cover is about 160px,
the margin between two cover should be about 20px. When the view is
resized, the cover width and the margin should make a few adjustment.
"""

# pylint: disable=unused-argument
import logging
import random
from datetime import date
from typing import TypeVar, Optional, List, cast, Union, TYPE_CHECKING

from PyQt6.QtCore import (
    QAbstractListModel,
    QModelIndex,
    Qt,
    QObject,
    QEvent,
    QRectF,
    QRect,
    QSize,
    QSortFilterProxyModel,
    pyqtSignal,
)
from PyQt6.QtGui import (
    QImage,
    QColor,
    QResizeEvent,
    QGuiApplication,
    QBrush,
    QPainter,
    QTextOption,
    QFontMetrics,
    QAction,
)
from PyQt6.QtWidgets import (
    QStyle,
    QAbstractItemDelegate,
    QListView,
    QFrame,
    QMenu,
)

from feeluown.i18n import t
from feeluown.utils import aio
from feeluown.library import AlbumModel, AlbumType, PlaylistModel, VideoModel
from feeluown.utils.reader import wrap, create_reader, Reader
from feeluown.library import reverse
from feeluown.gui.helpers import (
    ItemViewNoScrollMixin,
    resize_font,
    ReaderFetchMoreMixin,
    painter_save,
    secondary_text_color,
    fetch_cover_wrapper,
    random_solarized_color,
)
from feeluown.i18n import human_readable_number

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)
T = TypeVar("T")


COLORS = {
    "yellow": "#b58900",
    "orange": "#cb4b16",
    "red": "#dc322f",
    "magenta": "#d33682",
    "violet": "#6c71c4",
    "blue": "#268bd2",
    "cyan": "#2aa198",
    "green": "#859900",
}


class ImgCardListModel(QAbstractListModel, ReaderFetchMoreMixin[T]):
    def __init__(self, reader, fetch_image, source_name_map=None, parent=None):
        """

        :param reader: objects in reader should have `name` property
        :param fetch_image: func(item, cb, uid)
        :type reader: Iterable
        """
        super().__init__(parent)

        self.reader = self._reader = wrap(reader)
        self._fetch_more_step = 10
        self._items: List[T] = []
        self._is_fetching = False

        self._deleted_items: List[T] = []

        self.source_name_map = source_name_map or {}
        self.fetch_image = fetch_image
        self.colors = []
        self.images = {}  # {uri: QImage}

    @classmethod
    def create(cls, reader, app: "GuiApp"):
        return cls(
            create_reader(reader),
            fetch_image=fetch_cover_wrapper(app),
            source_name_map={p.identifier: p.name for p in app.library.list()},
        )

    def remove_item(self, item: T):
        """Remove an item from the model.

        .. versionadded: 5.0
        """
        try:
            row = self._items.index(item)
        except IndexError:
            return
        else:
            self.beginRemoveRows(QModelIndex(), row, row)
            self._items.remove(item)
            self._deleted_items.append(item)
            self.endRemoveRows()

    def can_fetch_more(self, _=None):
        """
        In order to handle the situation where items are removed from the model,
        the deleted items list should be considered. So override this method.
        """
        reader = cast(Reader, self._reader)
        count = reader.count
        if count is not None:
            return count > self.rowCount() + len(self._deleted_items)

        # The reader sets the count when it has no more items,
        # so it is safe to return True here
        return True

    def rowCount(self, _=QModelIndex()):
        return len(self._items)

    def _fetch_more_cb(self, items: Optional[List[T]]):
        self._is_fetching = False
        # None means an error occured.
        if items is None:
            return
        if items is not None and not items:
            self.no_more_item.emit()
            return
        items_len = len(items)
        colors = [random.choice(list(COLORS.values())) for _ in range(0, items_len)]
        self.colors.extend(colors)
        self.on_items_fetched(items)
        for item in items:
            aio.create_task(self.fetch_image(item, self._fetch_image_callback(item)))

    def _fetch_image_callback(self, item):
        def cb(content):
            uri = reverse(item)
            if content is None:
                self.images[uri] = None
                return

            img = QImage()
            img.loadFromData(content)
            self.images[uri] = img
            row = self._items.index(item)
            top_left = self.createIndex(row, 0)
            bottom_right = self.createIndex(row, 0)
            self.dataChanged.emit(top_left, bottom_right)

        return cb

    def data(self, index, role):
        offset = index.row()
        if not index.isValid() or offset >= len(self._items):
            return None
        item = self._items[offset]
        if role == Qt.ItemDataRole.DecorationRole:
            uri = reverse(item)
            image = self.images.get(uri)
            if image is not None:
                return image
            color_str = self.colors[offset]
            color = QColor(color_str)
            color.setAlphaF(0.8)
            return color
        elif role == Qt.ItemDataRole.DisplayRole:
            return item.name_display
        elif role == Qt.ItemDataRole.UserRole:
            return item
        elif role == Qt.ItemDataRole.WhatsThisRole:
            return self.source_name_map.get(item.source, item.source)
        elif role == Qt.ItemDataRole.ToolTipRole:
            return item.name
        return None


class ImgCardListDelegate(QAbstractItemDelegate):
    """Draw image card based on the view width and card_min_width.

    Card width is dynamically calculated based on the view width. The card width
    will be at least *card_min_width*.

    This delegate works well with ItemViewNoScrollMixin. It automatically adjusts
    the view's `_row_height` so that the view can show all the cards properly.
    Though changing the view's `_row_height` is tricky :)

    Card layout should be like the following::

        |card0    card1    card2|
                                          <- vertical spacing
        |card3    card4    card5|
                                          <- vertical spacing

    The leftmost cards should have a half_h_spacing on the right side,
    and the rightmost cards should have a half_h_spacing on the left side.
    Middle cards should have a half_h_spacing on both sides.
    """

    def __init__(
        self, parent=None, card_min_width=150, card_spacing=20, card_text_height=40, **_
    ):
        super().__init__(parent)

        self.view: "ImgCardListView" = parent
        self.view.installEventFilter(self)
        self.as_circle = True
        self.w_h_ratio = 1.0

        self._device_pixel_ratio = QGuiApplication.instance().devicePixelRatio()

        self.card_min_width = card_min_width
        self.card_spacing = card_spacing
        self.card_text_height = card_text_height

        self.h_spacing = self.card_spacing
        self.v_spacing = self.half_h_spacing = self.h_spacing // 2
        self.text_height = self.card_text_height

        # These variables are calculated in on_view_resized().
        self._card_width = self._card_height = 0
        self._view_width = 0

    def update_settings(self, name, value):
        assert hasattr(self, name), f"no such setting: {name}"
        setattr(self, name, value)
        self.re_calc_all()
        self.view.update()

    def paint(self, painter, option, index):
        obj: Optional[Union[QImage, QColor]] = index.data(
            Qt.ItemDataRole.DecorationRole
        )
        if obj is None:
            return

        with painter_save(painter):
            painter.setRenderHints(
                QPainter.RenderHint.Antialiasing
                | QPainter.RenderHint.SmoothPixmapTransform
                | QPainter.RenderHint.LosslessImageRendering
            )
            painter.translate(option.rect.x(), option.rect.y())

            if not self.is_leftmost(index):
                painter.translate(self.half_h_spacing, 0)

            spacing = self.get_card_h_spacing(index)
            draw_width = option.rect.width() - spacing

            secondary_color = border_color = secondary_text_color(option.palette)
            # Draw cover or color.
            img_height = int(draw_width / self.w_h_ratio)

            with painter_save(painter):
                self.draw_img_or_color(
                    painter, border_color, obj, draw_width, img_height
                )

            # Draw text(album name / artist name / playlist name), and draw source.
            text_title_height = 30
            text_source_height = self.text_height - text_title_height
            painter.translate(0, img_height)
            text_rect = QRectF(0, 0, draw_width, text_title_height)

            with painter_save(painter):
                # Draw hover effect.
                if option.state & QStyle.StateFlag.State_MouseOver:
                    painter.setPen(QColor(random_solarized_color()))
                self.draw_title(painter, index, text_rect)

            painter.translate(0, text_title_height - 5)
            with painter_save(painter):
                self.draw_whats_this(
                    painter,
                    index,
                    secondary_color,
                    QRectF(0, 0, draw_width, text_source_height + 5),
                )

    def draw_img_or_color(self, painter, border_color, obj, draw_width, height):
        pen = painter.pen()
        pen.setColor(border_color)
        painter.setPen(pen)
        if isinstance(obj, QColor):
            color = obj
            brush = QBrush(color)
            painter.setBrush(brush)
        else:
            img_w = obj.width()
            img_h = obj.height()
            if height <= 0 or img_w <= 0 or img_h <= 0:
                # Fall back to a flat fill when the image is invalid or height is zero.
                brush = QBrush(border_color)
            else:
                if img_w / img_h > draw_width / height:
                    img = obj.scaledToHeight(
                        int(height * self._device_pixel_ratio),
                        Qt.TransformationMode.SmoothTransformation,
                    )
                else:
                    img = obj.scaledToWidth(
                        int(draw_width * self._device_pixel_ratio),
                        Qt.TransformationMode.SmoothTransformation,
                    )
                img.setDevicePixelRatio(self._device_pixel_ratio)
                brush = QBrush(img)
            painter.setBrush(brush)
        border_radius = 3
        if self.as_circle:
            border_radius = draw_width // 2
        cover_rect = QRect(0, 0, draw_width, height)
        painter.drawRoundedRect(cover_rect, border_radius, border_radius)

    def draw_title(self, painter, index, text_rect):
        text_option = QTextOption()
        if self.as_circle:
            text_option.setAlignment(
                Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
            )
        else:
            text_option.setAlignment(
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
            )
        name = index.data(Qt.ItemDataRole.DisplayRole)
        fm = QFontMetrics(painter.font())
        elided_name = fm.elidedText(
            name, Qt.TextElideMode.ElideRight, int(text_rect.width())
        )
        painter.drawText(text_rect, elided_name, text_option)

    def draw_whats_this(self, painter, index, non_text_color, whats_this_rect):
        source_option = QTextOption()
        if self.as_circle:
            source_option.setAlignment(
                Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop
            )
        else:
            source_option.setAlignment(
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop
            )
        whats_this = index.data(Qt.ItemDataRole.WhatsThisRole)
        pen = painter.pen()
        font = painter.font()
        resize_font(font, -2)
        painter.setFont(font)
        pen.setColor(non_text_color)
        painter.setPen(non_text_color)
        painter.drawText(whats_this_rect, whats_this, source_option)

    def sizeHint(self, option, index):
        spacing = self.get_card_h_spacing(index)
        width = self._card_width
        if index.isValid():
            height = int(width / self.w_h_ratio) + self.text_height
            return QSize(width + spacing, height + self.v_spacing)
        return super().sizeHint(option, index)

    def on_view_resized(self, size: QSize, _: QSize):
        self._view_width = size.width()
        self.re_calc_all()
        # When view's height is changed, view.ajust_height() should do nothing.
        # So it does not cause infinite loop.
        self.view.adjust_height()

    def re_calc_all(self):
        # HELP: CardListView needs about 20 spacing left on macOS
        width = max(0, self._view_width - 20)
        card_spacing = self.card_spacing

        # according to our algorithm, when the widget width is:
        #   2(card_min_width + card_spacing) + card_spacing - 1,
        # the cover width can take the maximum width, it will be:
        #   CoverMaxWidth = 2 * card_min_width + card_spacing - 1

        # calculate max column count
        count = (width + card_spacing) // (self.card_min_width + card_spacing)
        count = max(count, 1)
        # calculate img_width when column count is the max
        self._card_width = (width + card_spacing) // count - card_spacing
        self._card_height = int(self._card_width / self.w_h_ratio) + self.text_height
        self.view.set_row_height(self._card_height + self.v_spacing)

    def column_count(self):
        return (self._view_width + self.card_spacing) // (
            self._card_width + self.card_spacing
        )

    def which_column(self, index: QModelIndex):
        if not self.view.isWrapping():
            return index.row()
        return index.row() % self.column_count()

    def which_row(self, index: QModelIndex):
        if not self.view.isWrapping():
            return 0
        return index.row() // self.column_count()

    def is_leftmost(self, index):
        return self.which_column(index) == 0

    def is_rightmost(self, index):
        if self.view.isWrapping():
            return self.which_column(index) == self.column_count() - 1
        return False  # HELP: no way to check if it is the rightmost.

    def get_card_h_spacing(self, index):
        if self.is_leftmost(index) or self.is_rightmost(index):
            return self.half_h_spacing
        return self.h_spacing

    def eventFilter(self, _: QObject, event: QEvent):
        if event.type() == QEvent.Type.Resize:
            event = cast(QResizeEvent, event)
            self.on_view_resized(event.size(), event.oldSize())
        return False


class ImgFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent=None, types=None):
        super().__init__(parent)

        self.text = ""

    def filter_by_text(self, text):
        if text == self.text:
            return
        self.text = text
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        accepted = True
        source_model = cast(ImgCardListModel, self.sourceModel())
        index = source_model.index(source_row, parent=source_parent)
        artist = index.data(Qt.ItemDataRole.UserRole)
        if self.text:
            accepted = self.text.lower() in artist.name_display.lower()
        return accepted


class ImgCardListView(ItemViewNoScrollMixin, QListView):
    """
    .. versionchanged:: 3.9
       The *card_min_width*, *card_spacing*, *card_text_height* parameter were removed.
    """

    # Generic signal emitted when a remove action is requested from context menu.
    # .. versionadded:: 5.0
    #
    # NOTE: Compared to SongsTableView.remove_song_func implementation,
    # this way does not block the UI when removing an item.
    # At the same time, it provides a callback to notify the success status.
    # I'm not sure whether the design of the callback signature is good.
    remove_item_needed = pyqtSignal([object, object])
    remove_action_text = t("remove-action")

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

        # Override ItemViewNoScrollMixin variables. Actually, there variables are
        # not important because ItemViewNoScrollMixin use QListView.sizeHint() to
        # calculate the size and it works well.
        self._least_row_count = 1
        self._row_height = 0

        self.setViewMode(QListView.ViewMode.IconMode)
        self.setResizeMode(QListView.ResizeMode.Adjust)
        self.setWrapping(True)
        self.setMouseTracking(True)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.initialize()

        self.activated.connect(self.on_activated)

        # Text for the remove action in context menu. Subclasses can override.
        self.enable_remove_action = False

    def on_activated(self, _: QModelIndex):
        """
        Subclass can implement this method if needed.
        """
        pass

    def contextMenuEvent(self, event):
        """Generic context menu that provides a remove action when available.

        Subclasses may set `remove_action_text` to a non-empty string to enable
        the remove action. When selected, `remove_item_needed` will be emitted
        with the item (from UserRole) at the clicked index.
        """
        if not self.enable_remove_action:
            return

        index = self.indexAt(event.pos())
        if not index.isValid():
            return

        # If no remove action text is provided, do nothing (no menu by default).
        item = index.data(Qt.ItemDataRole.UserRole)
        menu = QMenu(self)
        remove_action = QAction(self.remove_action_text, self)
        remove_action.triggered.connect(
            lambda: self.remove_item_needed.emit(item, self.remove_item_cb)
        )
        menu.addAction(remove_action)
        menu.exec(event.globalPos())

    def remove_item_cb(self, item, success: bool):
        logger.info(f"remove item({item}) cb {'ok' if success else 'fail'}")
        if success:
            model = self.model()
            if isinstance(model, QSortFilterProxyModel):
                model = model.sourceModel()
            model.remove_item(item)


class VideoCardListModel(ImgCardListModel):
    def data(self, index, role):
        offset = index.row()
        if not index.isValid() or offset >= len(self._items):
            return None
        video = self._items[offset]
        if role == Qt.ItemDataRole.DisplayRole:
            return video.title
        elif role == Qt.ItemDataRole.ToolTipRole:
            return video.title
        elif (
            role == Qt.ItemDataRole.WhatsThisRole
            and isinstance(video, VideoModel)
            and video.play_count > 0
        ):
            return f"▶ {human_readable_number(video.play_count)}"
        return super().data(index, role)


class VideoCardListDelegate(ImgCardListDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.as_circle = False
        self.w_h_ratio = 1.618


class VideoFilterProxyModel(ImgFilterProxyModel):
    pass


class VideoCardListView(ImgCardListView):
    play_video_needed = pyqtSignal([object])
    # Use base class remove behavior; subclasses keep a convenient alias
    remove_action_text = t("remove-action-video")

    def on_activated(self, index):
        video = index.data(Qt.ItemDataRole.UserRole)
        self.play_video_needed.emit(video)


class PlaylistCardListModel(ImgCardListModel):
    def data(self, index, role):
        offset = index.row()
        if not index.isValid() or offset >= len(self._items):
            return None

        playlist = self._items[offset]
        if (
            role == Qt.ItemDataRole.WhatsThisRole
            and isinstance(playlist, PlaylistModel)
            and playlist.play_count > 0
        ):
            count = human_readable_number(playlist.play_count)
            return f"▶ {count}"
        return super().data(index, role)


class PlaylistCardListDelegate(ImgCardListDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.as_circle = False


class PlaylistFilterProxyModel(ImgFilterProxyModel):
    pass


class PlaylistCardListView(ImgCardListView):
    show_playlist_needed = pyqtSignal([object])
    remove_action_text = t("remove-action-playlist")

    def on_activated(self, index):
        artist = index.data(Qt.ItemDataRole.UserRole)
        self.show_playlist_needed.emit(artist)


class ArtistCardListModel(ImgCardListModel):
    pass


class ArtistCardListDelegate(ImgCardListDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.as_circle = True


class ArtistFilterProxyModel(ImgFilterProxyModel):
    pass


class ArtistCardListView(ImgCardListView):
    show_artist_needed = pyqtSignal([object])
    remove_action_text = t("remove-action-musician")

    def on_activated(self, index):
        artist = index.data(Qt.ItemDataRole.UserRole)
        self.show_artist_needed.emit(artist)


class AlbumCardListModel(ImgCardListModel):
    def data(self, index, role):
        offset = index.row()
        if not index.isValid() or offset >= len(self._items):
            return None

        album = self._items[offset]
        if role == Qt.ItemDataRole.WhatsThisRole:
            if isinstance(album, AlbumModel):
                try:
                    year, month, day = map(int, album.released.split("-"))
                except Exception:
                    year, month, day = 1970, 1, 1
                release_date = date(year, month, day)
                return t(
                    "album-release-date",
                    releaseDate=release_date,
                    trackCount=album.song_count,
                )
        return super().data(index, role)


class AlbumCardListDelegate(ImgCardListDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.as_circle = False


class AlbumFilterProxyModel(ImgFilterProxyModel):
    def __init__(self, parent=None, types=None):
        super().__init__(parent)

        self.types = types

    def filter_by_types(self, types):
        # if types is a empty list or None, we show all albums
        if not types:
            types = None
        self.types = types
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        accepted = super().filterAcceptsRow(source_row, source_parent)
        source_model = self.sourceModel()
        assert isinstance(source_model, AlbumCardListModel)
        index = source_model.index(source_row, parent=source_parent)
        album = index.data(Qt.ItemDataRole.UserRole)
        if accepted and self.types:
            accepted = AlbumType(album.type_) in self.types
        return accepted


class AlbumCardListView(ImgCardListView):
    show_album_needed = pyqtSignal([object])
    remove_action_text = t("remove-action-album")

    def on_activated(self, index):
        album = index.data(Qt.ItemDataRole.UserRole)
        self.show_album_needed.emit(album)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from feeluown.library import Library

from fuo_netease import provider as np

logging.basicConfig()
logger = logging.getLogger('feeluown')
logger.setLevel(logging.DEBUG)

lib = Library()
lib.register(np)

library = lib



import sys
from typing import Optional

from PyQt6.QtCore import QTimer, QEvent, Qt, QRect
from PyQt6.QtWidgets import (
    QWidget,
    QFrame,
    QHBoxLayout,
    QVBoxLayout,
    QPushButton,
)
from PyQt6.QtGui import (
    QMouseEvent,
    QCursor,
    QPainter,
    QPalette,
    QBrush,
)

from feeluown.gui.helpers import darker_or_lighter
from feeluown.gui.widgets.cover_label import CoverLabelV2
from feeluown.gui.widgets.progress_slider import ProgressSlider
from feeluown.gui.components import (
    LineSongLabel,
    MediaButtonsV2,
    LyricButton,
    WatchButton,
    LikeButton,
    MVButton,
    VolumeSlider,
    SongSourceTag,
    PlayerProgressRatioLabel,
)

IS_MACOS = sys.platform == "darwin"


class MouseState:
    def __init__(self, e: QMouseEvent):
        self.start_pos = e.globalPosition()
        self.current_pos = self.start_pos

    @property
    def moved(self) -> bool:
        return self.start_pos != self.current_pos


class Toolbar(QWidget):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

        button_width = 20
        self._song_btn_size = (16, 16)
        # self.setStyleSheet('border: 1px solid red;')

        self.song_source_tag = SongSourceTag(app=self._app, font_size=10)
        self.line_song_label = LineSongLabel(app=self._app)
        self.progress_slider = ProgressSlider(app=self._app)
        self.progress_label = PlayerProgressRatioLabel(app=self._app)
        self.volume_slider = VolumeSlider(app=self._app)
        self.media_buttons = MediaButtonsV2(
            app=self._app, spacing=0, button_width=button_width + 5
        )
        self.lyric_button = LyricButton(app=self._app)
        self.like_button = LikeButton(app=self._app, size=self._song_btn_size)
        self.watch_button = WatchButton(app=self._app)
        self.mv_button = MVButton(app=self._app, height=self._song_btn_size[1])
        self.volume_button = QPushButton()

        self.volume_button.setObjectName("volume_btn")
        self.volume_button.setFixedWidth(button_width)
        self.volume_slider.setMaximumWidth(60)

        # Set margins/spacing explicitly because different platforms
        # has different default values.
        self._layout = QHBoxLayout(self)
        self._layout.setSpacing(6)
        bottom_margin = 0 if IS_MACOS else 8
        self._layout.setContentsMargins(8, 8, 8, bottom_margin)

        self._v_layout = QVBoxLayout()
        self._song_layout = QHBoxLayout()
        self._btns_layout = QHBoxLayout()
        self._other_btns_layout = QVBoxLayout()

        self._layout.addLayout(self._v_layout)
        self._layout.addLayout(self._other_btns_layout)

        self._v_layout.setSpacing(0)
        self._v_layout.addLayout(self._song_layout)
        self._v_layout.addSpacing(3)
        self._v_layout.addStretch(0)
        self._v_layout.addWidget(self.progress_slider)
        self._v_layout.addStretch(0)
        self._v_layout.addLayout(self._btns_layout)

        if IS_MACOS:
            # On macOS, the default height of the slider is not enough to show
            # the slider handler, so set a minimum height for sliders.
            self.progress_slider.setMinimumHeight(24)
            self.volume_slider.setMinimumHeight(24)

        self._song_layout.setSpacing(self._song_btn_size[1] // 2)
        self._song_layout.addWidget(self.song_source_tag)
        self._song_layout.addWidget(self.line_song_label)
        self._song_layout.addWidget(self.mv_button)
        self._song_layout.addWidget(self.like_button)

        self._other_btns_layout.addWidget(self.watch_button)
        self._other_btns_layout.addWidget(self.lyric_button)
        self.watch_button.hide()
        self.lyric_button.hide()

        self._btns_layout.setSpacing(0)
        self._btns_layout.addWidget(self.progress_label)
        self._btns_layout.addStretch(0)
        self._btns_layout.addWidget(self.media_buttons)
        self._btns_layout.addStretch(0)
        self._btns_layout.addWidget(self.volume_button)
        self._btns_layout.addWidget(self.volume_slider)


class AnimatedCoverLabel(CoverLabelV2):
    def __init__(self, app, padding=6, *args, **kwargs):
        super().__init__(app, *args, **kwargs)

        self._padding = padding
        self._angle: float = 0
        self._timer = QTimer()
        self._timer.timeout.connect(self.on_timeout)
        self._timer.start(16)

    def on_timeout(self):
        self._angle += 0.2
        self.update()

    def paintEvent(self, e):
        radius = self._radius
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)

        center = (self.width() // 2, self.height() // 2)

        painter.save()
        painter.translate(*center)
        painter.rotate(self._angle)
        painter.translate(-center[0], -center[1])
        painter.setPen(Qt.PenStyle.NoPen)

        # Draw border.
        color = darker_or_lighter(self.palette().color(QPalette.ColorRole.Window), 115)
        painter.setBrush(QBrush(color))
        painter.drawRoundedRect(self.rect(), radius, radius)

        pixmap = self.drawer.get_pixmap()
        if pixmap is not None:
            size = pixmap.size()
            y = (size.height() - self.height()) // 2
            rect = QRect(
                self._padding,
                y + self._padding,
                self.width() - self._padding * 2,
                self.height() - self._padding * 2,
            )
            brush = QBrush(pixmap)
            painter.setBrush(brush)
            painter.drawRoundedRect(rect, radius, radius)

        painter.restore()
        painter.end()


class FloatingBox(QFrame):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._height = 100
        self._padding = 6

        self._app = app
        self.setMouseTracking(True)

        self._timer = QTimer(self)

        # Windows movement control.
        self._mouse_state: Optional[MouseState] = None

        self.toolbar = Toolbar(app=self._app)
        self.circle = AnimatedCoverLabel(
            app=self._app, radius=self._height // 2, padding=self._padding
        )
        self.toolbar.installEventFilter(self)
        self.circle.installEventFilter(self)
        self.circle.setMouseTracking(True)

        self.circle.setFixedSize(self._height, self._height)
        self.toolbar.setFixedSize(300, self._height)
        self.setFixedHeight(self._height)

        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.toolbar)
        self._layout.addWidget(self.circle)
        self.maybe_hide_toolbar()

        self._timer.timeout.connect(self.maybe_hide_toolbar)

    def paintEvent(self, e):
        super().paintEvent(e)

        if not self.toolbar.isVisible():
            return

        # Draw background for toolbar.
        new_bg_color = darker_or_lighter(
            self.palette().color(QPalette.ColorRole.Window), 115
        )
        painter = QPainter(self)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.save()
        painter.setBrush(new_bg_color)
        painter.drawRoundedRect(
            QRect(0, 0, self.width() - self._height // 2, self._height), 3, 3
        )
        # painter.drawEllipse(self.circle.rect())
        painter.restore()

    async def show_cover(self, *args, **kwargs):
        await self.circle.show_cover(*args, **kwargs)

    def maybe_show_toolbar(self):
        if not self.toolbar.isVisible():
            self.toolbar.show()
            self.setFixedWidth(self.circle.width() + self.toolbar.width())
            self.move(self.pos().x() - self.toolbar.width(), self.pos().y())

    def maybe_hide_toolbar(self):
        if self.toolbar.isVisible():
            pos = self.toolbar.mapFromGlobal(QCursor.pos())
            if self.rect().contains(pos):
                return
        self.toolbar.hide()
        self.move(self.pos().x() + self.toolbar.width(), self.pos().y())
        self._timer.stop()
        self.setFixedWidth(self.circle.width())

    def mousePressEvent(self, e):
        self._mouse_state = MouseState(e)
        super().mousePressEvent(e)

    def mouseMoveEvent(self, e: QMouseEvent):
        self.maybe_show_toolbar()
        self._timer.start(1000)

        # NOTE: e.button() == Qt.MouseButton.LeftButton don't work on Windows
        # on Windows, even I drag with LeftButton, the e.button() return 0,
        # which means no button
        if self._mouse_state is not None:
            delta = e.globalPosition() - self._mouse_state.current_pos
            self.move(int(self.x() + delta.x()), int(self.y() + delta.y()))
            self._mouse_state.current_pos = e.globalPosition()
        super().mouseMoveEvent(e)

    def mouseReleaseEvent(self, e):
        # If the window is moved, intercept the event.
        moved = self._mouse_state is not None and self._mouse_state.moved
        self._mouse_state = None
        if not moved:
            super().mouseReleaseEvent(e)

    def eventFilter(self, obj, event):
        if obj in self.children() and event.type() == QEvent.Type.MouseMove:
            self.maybe_show_toolbar()
            self._timer.start(1000)
            return False
        return False


if __name__ == "__main__":
    import os

    from PyQt6.QtWidgets import QApplication
    from PyQt6.QtCore import QDir
    from PyQt6.QtGui import QImage
    from unittest.mock import MagicMock

    from feeluown.gui.theme import ThemeManager

    icons_dir = os.path.join("feeluown", "gui/assets/icons")
    QDir.addSearchPath("icons", icons_dir)

    img = QImage()
    # !!! You should change the image filename.
    img_fn = "7c90bb4edfa99cae1d142a33ebe26673-1685249600"
    img_fp = os.path.expanduser(f"~/.FeelUOwn/cache/{img_fn}")
    with open(img_fp, "rb") as f:
        img.loadFromData(f.read())

    app = MagicMock()
    qapp = QApplication([])
    theme_mgr = ThemeManager(app)
    box = FloatingBox(app)
    box.setWindowFlags(Qt.WindowType.FramelessWindowHint)
    box.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
    box.circle.show_img(img)
    box.toolbar.line_song_label.setText("Haha - hehe")
    box.show()
    box.move(600, 400)
    theme_mgr.load_light()
    qapp.exec()



import asyncio
import os
from contextlib import contextmanager

from PyQt6.QtCore import QDir
from PyQt6.QtWidgets import QApplication, QWidget, QHBoxLayout

from feeluown.utils.compat import DefaultQEventLoopPolicy
from feeluown.debug import *  # noqa


@contextmanager
def simple_qapp():
    qapp = QApplication([])
    yield qapp
    qapp.exec()


@contextmanager
def async_simple_qapp():
    app_close_event = asyncio.Event()
    qapp = QApplication([])
    qapp.aboutToQuit.connect(app_close_event.set)
    yield qapp
    asyncio.set_event_loop_policy(DefaultQEventLoopPolicy())
    asyncio.run(app_close_event.wait())


def read_dark_theme_qss():
    from feeluown.gui.theme import read_resource

    pkg_root_dir = os.path.join(os.path.dirname(__file__), "..")
    icons_dir = os.path.join(pkg_root_dir, "gui/assets/icons")
    QDir.addSearchPath("icons", icons_dir)

    qss = read_resource("common.qss")
    dark = read_resource("dark.qss")
    return qss + "\n" + dark


@contextmanager
def simple_layout(cls=QHBoxLayout, theme="", aio=False):
    func = async_simple_qapp if aio is True else simple_qapp
    with func():
        main = QWidget()
        if theme == "dark":
            main.setStyleSheet(read_dark_theme_qss())
        layout = cls(main)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        yield layout
        main.show()



"""
This example proves the following experience (under multiple inheritance circumstances)

1. The `Base.initialize` method is only called once with `super().initialize()`
"""
class Base:
    def __init__(self):
        print('base init')

    def initialize(self):
        print('base initialize')

    def run(self):
        print('base run')


class Widget:
    def __init__(self, parent):
        print('widget init')


class A(Base):
    def __init__(self):
        super().__init__()
        print('A init')

    def initialize(self):
        super().initialize()
        print('a initialize')

    def run(self):
        super().run()
        print('a run')


class B(Base, Widget):
    def __init__(self):
        Base.__init__(self)
        Widget.__init__(self, self)
        print('B init')

    def initialize(self):
        super().initialize()
        print('b initialize')

    def run(self):
        super().run()
        print('b run')


class M(A, B):
    def __init__(self):
        super().__init__()

    def initialize(self):
        super().initialize()
        print('mixed initialize')

    def run(self):
        super().run()
        print('m run')


m = M()
m.initialize()
m.run()
